# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from textwrap import dedent
from unittest.mock import ANY, call, patch

import fixtures
from testtools.matchers import Contains, Equals, FileContains, Not

from snapcraft.internal.build_providers._snap import (
    SnapInjector,
    _get_snap_channel,
    repo,
)
from tests import fixture_setup, unit

from . import ProviderImpl, get_project


class SnapInjectionTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        patcher = patch("snapcraft.internal.repo.snaps.get_assertion")
        self.get_assertion_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.registry_filepath = os.path.join(self.path, "registry.yaml")

        self.provider = ProviderImpl(project=get_project(), echoer=lambda x: x)

    def test_snapcraft_installed_on_host_from_store(self):
        self.fake_snapd.snaps_result = [
            {
                "name": "snapd",
                "confinement": "strict",
                "id": "2kkitQ",
                "channel": "edge",
                "revision": "1",
                "tracking-channel": "latest/edge",
            },
            {
                "name": "core18",
                "confinement": "strict",
                "id": "2kkibb",
                "channel": "stable",
                "revision": "123",
                "tracking-channel": "latest/beta",
            },
            {
                "name": "snapcraft",
                "confinement": "classic",
                "id": "3lljuR",
                "channel": "edge",
                "revision": "345",
                "tracking-channel": "latest/candidate",
            },
        ]
        self.get_assertion_mock.side_effect = [
            b"fake-assertion-account-store",
            b"fake-assertion-declaration-snapd",
            b"fake-assertion-revision-snapd-1",
            b"fake-assertion-account-store",
            b"fake-assertion-declaration-core18",
            b"fake-assertion-revision-core18-123",
            b"fake-assertion-account-store",
            b"fake-assertion-declaration-snapcraft",
            b"fake-assertion-revision-snapcraft-345",
        ]

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("snapd")
        snap_injector.add("core18")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        get_assertion_calls = [
            call(
                [
                    "account-key",
                    "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
                ]
            ),
            call(["snap-declaration", "snap-name=snapd"]),
            call(["snap-revision", "snap-revision=1", "snap-id=2kkitQ"]),
            call(
                [
                    "account-key",
                    "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
                ]
            ),
            call(["snap-declaration", "snap-name=core18"]),
            call(["snap-revision", "snap-revision=123", "snap-id=2kkibb"]),
            call(
                [
                    "account-key",
                    "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
                ]
            ),
            call(["snap-declaration", "snap-name=snapcraft"]),
            call(["snap-revision", "snap-revision=345", "snap-id=3lljuR"]),
        ]
        self.get_assertion_mock.assert_has_calls(get_assertion_calls)
        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", "experimental.snapd-snap=true"]),
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "ack", "/var/tmp/snapd.assert"]),
                call(["snap", "install", "/var/tmp/snapd.snap"]),
                call(["snap", "switch", "snapd", "--channel", "latest/edge"]),
                call(["snap", "ack", "/var/tmp/core18.assert"]),
                call(["snap", "install", "/var/tmp/core18.snap"]),
                call(["snap", "switch", "core18", "--channel", "latest/beta"]),
                call(["snap", "ack", "/var/tmp/snapcraft.assert"]),
                call(["snap", "install", "--classic", "/var/tmp/snapcraft.snap"]),
                call(["snap", "switch", "snapcraft", "--channel", "latest/candidate"]),
            ]
        )
        self.provider.push_file_mock.assert_has_calls(
            [
                call(source=ANY, destination="/var/tmp/snapd.snap"),
                call(source=ANY, destination="/var/tmp/snapd.assert"),
                call(source=ANY, destination="/var/tmp/core18.snap"),
                call(source=ANY, destination="/var/tmp/core18.assert"),
                call(source=ANY, destination="/var/tmp/snapcraft.snap"),
                call(source=ANY, destination="/var/tmp/snapcraft.assert"),
            ]
        )
        self.assertThat(
            self.registry_filepath,
            FileContains(
                dedent(
                    """\
                    core18:
                    - revision: '123'
                    snapcraft:
                    - revision: '345'
                    snapd:
                    - revision: '1'
                    """
                )
            ),
        )

    def test_snapcraft_installed_on_host_from_store_but_injection_disabled(self):
        self.useFixture(fixture_setup.FakeStore())

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
            inject_from_host=False,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        self.get_assertion_mock.assert_not_called()
        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "install", "--channel", "stable", "core"]),
                call(
                    ["snap", "install", "--classic", "--channel", "stable", "snapcraft"]
                ),
            ]
        )
        self.provider.push_file_mock.assert_not_called()
        self.assertThat(
            self.registry_filepath,
            FileContains(
                dedent(
                    """\
                    core:
                    - revision: '10000'
                    snapcraft:
                    - revision: '25'
                    """
                )
            ),
        )

    def test_snapcraft_installed_on_host_with_dangerous(self):
        self.fake_snapd.snaps_result = [
            {
                "name": "core",
                "confinement": "strict",
                "id": "2kkitQ",
                "channel": "stable",
                "revision": "123",
            },
            {
                "name": "snapcraft",
                "confinement": "classic",
                "revision": "x20",
                "tracking-channel": "latest/stable",
            },
        ]
        self.get_assertion_mock.side_effect = [
            b"fake-assertion-account-store",
            b"fake-assertion-declaration-core",
            b"fake-assertion-revision-core-123",
        ]

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        get_assertion_calls = [
            call(
                [
                    "account-key",
                    "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
                ]
            ),
            call(["snap-declaration", "snap-name=core"]),
            call(["snap-revision", "snap-revision=123", "snap-id=2kkitQ"]),
        ]
        self.get_assertion_mock.assert_has_calls(get_assertion_calls)
        # Check the call count to ensure the snap switch command does not sneak in.
        self.assertThat(self.provider.run_mock.call_count, Equals(6))
        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "ack", "/var/tmp/core.assert"]),
                call(["snap", "install", "/var/tmp/core.snap"]),
                call(["snap", "ack", "/var/tmp/snapcraft.assert"]),
                call(
                    [
                        "snap",
                        "install",
                        "--dangerous",
                        "--classic",
                        "/var/tmp/snapcraft.snap",
                    ]
                ),
            ]
        )
        self.assertThat(
            self.registry_filepath,
            FileContains(
                dedent(
                    """\
                    core:
                    - revision: '123'
                    snapcraft:
                    - revision: x20
                    """
                )
            ),
        )
        self.provider.push_file_mock.assert_has_calls(
            [
                call(source=ANY, destination="/var/tmp/core.snap"),
                call(source=ANY, destination="/var/tmp/core.assert"),
                call(source=ANY, destination="/var/tmp/snapcraft.snap"),
                call(source=ANY, destination="/var/tmp/snapcraft.assert"),
            ]
        )

    def test_snapcraft_not_installed_on_host(self):
        self.useFixture(fixture_setup.FakeStore())

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        self.get_assertion_mock.assert_not_called()
        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "install", "--channel", "stable", "core"]),
                call(
                    ["snap", "install", "--classic", "--channel", "stable", "snapcraft"]
                ),
            ]
        )
        self.provider.push_file_mock.assert_not_called()
        self.assertThat(
            self.registry_filepath,
            FileContains(
                dedent(
                    """\
                    core:
                    - revision: '10000'
                    snapcraft:
                    - revision: '25'
                    """
                )
            ),
        )

    def test_snapcraft_not_installed_on_host_with_channel_from_environment(self):
        self.useFixture(fixture_setup.FakeStore())
        self.useFixture(
            fixtures.EnvironmentVariable(
                "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT", "latest/edge"
            )
        )

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        self.get_assertion_mock.assert_not_called()
        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "install", "--channel", "stable", "core"]),
                call(
                    ["snap", "install", "--classic", "--channel", "edge", "snapcraft"]
                ),
            ]
        )
        self.provider.push_file_mock.assert_not_called()
        self.assertThat(
            self.registry_filepath,
            FileContains(
                dedent(
                    """\
                    core:
                    - revision: '10000'
                    snapcraft:
                    - revision: '25'
                    """
                )
            ),
        )

    def test_no_registry(self):
        self.useFixture(fixture_setup.FakeStore())

        snap_injector = SnapInjector(
            registry_filepath=None,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "install", "--channel", "stable", "core"]),
                call(
                    ["snap", "install", "--classic", "--channel", "stable", "snapcraft"]
                ),
            ]
        )
        self.provider.run_mock.reset_mock()

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "install", "--channel", "stable", "core"]),
                call(
                    ["snap", "install", "--classic", "--channel", "stable", "snapcraft"]
                ),
            ]
        )

    def test_snapcraft_installed_on_host_from_store_rerun_is_nop(self):
        self.fake_snapd.snaps_result = [
            {
                "name": "core",
                "confinement": "strict",
                "id": "2kkitQ",
                "channel": "stable",
                "revision": "123",
            },
            {
                "name": "snapcraft",
                "confinement": "classic",
                "id": "3lljuR",
                "channel": "edge",
                "revision": "345",
            },
        ]
        self.get_assertion_mock.side_effect = [
            b"fake-assertion-account-store",
            b"fake-assertion-declaration-core",
            b"fake-assertion-revision-core-123",
            b"fake-assertion-account-store",
            b"fake-assertion-declaration-snapcraft",
            b"fake-assertion-revision-snapcraft-345",
        ]

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()
        self.provider.run_mock.reset_mock()

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        self.provider.run_mock.assert_not_called()

    def test_snapcraft_installed_on_host_from_store_rerun_refreshes(self):
        self.useFixture(fixture_setup.FakeStore())

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()
        self.provider.run_mock.reset_mock()

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")
        snap_injector.apply()

        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "refresh", "--channel", "stable", "core"]),
                call(
                    ["snap", "refresh", "--classic", "--channel", "stable", "snapcraft"]
                ),
            ]
        )

    def test_snapd_not_on_host_installs_from_store(self):
        self.useFixture(fixture_setup.FakeStore())

        snap_injector = SnapInjector(
            registry_filepath=self.registry_filepath,
            runner=self.provider._run,
            file_pusher=self.provider._push_file,
        )
        snap_injector.add("core")
        snap_injector.add("snapcraft")

        with patch(
            "snapcraft.internal.repo.snaps.SnapPackage.get_local_snap_info",
            side_effect=repo.errors.SnapdConnectionError("core", "url"),
        ):
            snap_injector.apply()

        self.get_assertion_mock.assert_not_called()
        self.provider.run_mock.assert_has_calls(
            [
                call(["snap", "set", "system", ANY]),
                call(["snap", "watch", "--last=auto-refresh?"]),
                call(["snap", "install", "--channel", "stable", "core"]),
                call(
                    ["snap", "install", "--classic", "--channel", "stable", "snapcraft"]
                ),
            ]
        )
        self.provider.push_file_mock.assert_not_called()
        self.assertThat(
            self.registry_filepath,
            FileContains(
                dedent(
                    """\
                    core:
                    - revision: '10000'
                    snapcraft:
                    - revision: '25'
                    """
                )
            ),
        )


class GetChannelTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

    def test_default_channel_for_snapcraft(self):
        self.assertThat(str(_get_snap_channel("snapcraft")), Equals("latest/stable"))
        self.assertThat(
            self.fake_logger.output,
            Not(
                Contains(
                    "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT is set: installing "
                    "snapcraft from"
                )
            ),
        )

    def test_channel_for_snapcraft_from_env(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT", "latest/edge"
            )
        )
        self.assertThat(str(_get_snap_channel("snapcraft")), Equals("latest/edge"))
        self.assertThat(
            self.fake_logger.output,
            Contains(
                "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT is set: installing "
                "snapcraft from latest/edge"
            ),
        )

    def test_default_channel_for_other_snap(self):
        self.assertThat(str(_get_snap_channel("core")), Equals("latest/stable"))
        self.assertThat(
            self.fake_logger.output,
            Not(
                Contains(
                    "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT is set: installing "
                    "snapcraft from"
                )
            ),
        )

    def test_channel_for_other_snap_not_affected_by_env(self):
        self.useFixture(
            fixtures.EnvironmentVariable(
                "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT", "latest/edge"
            )
        )
        self.assertThat(str(_get_snap_channel("core")), Equals("latest/stable"))
        self.assertThat(
            self.fake_logger.output,
            Not(
                Contains(
                    "SNAPCRAFT_BUILD_ENVIRONMENT_CHANNEL_SNAPCRAFT is set: installing "
                    "snapcraft from"
                )
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from snapcraft.internal.sources import errors


class TestErrorFormatting:

    scenarios = (
        (
            "SnapcraftSourceNotFoundError",
            {
                "exception_class": errors.SnapcraftSourceNotFoundError,
                "kwargs": {"source": "file-not-found.tar.gz"},
                "expected_message": (
                    "Failed to pull source: 'file-not-found.tar.gz'.\n"
                    "Please ensure the source path is correct and that it is accessible.\n"
                    "See `snapcraft help sources` for more information."
                ),
            },
        ),
        (
            "SnapcraftSourceUnhandledError",
            {
                "exception_class": errors.SnapcraftSourceUnhandledError,
                "kwargs": {"source": "unknown://source/type"},
                "expected_message": (
                    "Failed to pull source: "
                    "unable to determine source type of 'unknown://source/type'.\n"
                    "Check that the URL is correct or consider specifying "
                    "`source-type` for this part. "
                    "See `snapcraft help sources` for more information."
                ),
            },
        ),
        (
            "SnapcraftSourceInvalidOptionError",
            {
                "exception_class": errors.SnapcraftSourceInvalidOptionError,
                "kwargs": {"source_type": "test", "option": "foo"},
                "expected_message": (
                    "Failed to pull source: "
                    "'foo' cannot be used with a test source.\n"
                    "See `snapcraft help sources` for more information."
                ),
            },
        ),
        (
            "SnapcraftSourceIncompatibleOptionsError, two options",
            {
                "exception_class": errors.SnapcraftSourceIncompatibleOptionsError,
                "kwargs": {"source_type": "test", "options": ["foo", "bar"]},
                "expected_message": (
                    "Failed to pull source: "
                    "cannot specify both 'bar' and 'foo' for a test source.\n"
                    "See `snapcraft help sources` for more information."
                ),
            },
        ),
        (
            "InvalidSnapError",
            {
                "exception_class": errors.InvalidSnapError,
                "kwargs": {},
                "expected_message": (
                    "The snap file does not contain valid data. "
                    "Ensure the source lists a proper snap file"
                ),
            },
        ),
    )

    def test_error_formatting(self, exception_class, kwargs, expected_message):
        assert str(exception_class(**kwargs)) == expected_message


class TestSnapcraftException:

    scenarios = (
        (
            "GitCommandError",
            {
                "exception_class": errors.GitCommandError,
                "kwargs": dict(
                    command=["echo", "hi", "$foo"],
                    output="some error output",
                    exit_code=0,
                ),
                "expected_brief": "Failed to execute git command: echo hi '$foo'",
                "expected_resolution": "Consider checking your git configuration for settings which may cause issues.",
                "expected_details": "Command failed with exit code 0 and output:\nsome error output",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
    )

    def test_snapcraft_exception_handling(
        self,
        exception_class,
        kwargs,
        expected_brief,
        expected_resolution,
        expected_details,
        expected_docs_url,
        expected_reportable,
    ):
        exception = exception_class(**kwargs)
        assert exception.get_brief() == expected_brief
        assert exception.get_resolution() == expected_resolution
        assert exception.get_details() == expected_details
        assert exception.get_docs_url() == expected_docs_url
        assert exception.get_reportable() == expected_reportable



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import FileExists, Not

from tests import integration


class ScriptletTestCase(integration.TestCase):
    def test_override_pull(self):
        self.run_snapcraft("pull", "scriptlet-override-pull")

        sourcedir = os.path.join(self.parts_dir, "override-pull-scriptlet-test", "src")

        self.assertThat(os.path.join(sourcedir, "file"), FileExists())
        self.assertThat(os.path.join(sourcedir, "before-pull"), FileExists())
        self.assertThat(os.path.join(sourcedir, "after-pull"), FileExists())

    def test_override_build(self):
        self.run_snapcraft("build", "scriptlet-override-build")

        builddir = os.path.join(
            self.parts_dir, "override-build-scriptlet-test", "build"
        )
        installdir = os.path.join(
            self.parts_dir, "override-build-scriptlet-test", "install"
        )

        self.assertThat(os.path.join(builddir, "file"), FileExists())
        self.assertThat(os.path.join(builddir, "before-build"), FileExists())
        self.assertThat(os.path.join(builddir, "after-build"), FileExists())
        self.assertThat(os.path.join(installdir, "file"), FileExists())
        self.assertThat(os.path.join(installdir, "before-install"), FileExists())
        self.assertThat(os.path.join(installdir, "after-install"), FileExists())

    def test_override_stage(self):
        self.run_snapcraft("stage", "scriptlet-override-stage")

        installdir = os.path.join(
            self.parts_dir, "override-stage-scriptlet-test", "install"
        )

        # Assert that the before/after stage files aren't placed in the
        # installdir, although the file is.
        self.assertThat(os.path.join(installdir, "file"), FileExists())
        self.assertThat(os.path.join(installdir, "before-stage"), Not(FileExists()))
        self.assertThat(os.path.join(installdir, "after-stage"), Not(FileExists()))

        self.assertThat(os.path.join(self.stage_dir, "file"), FileExists())
        self.assertThat(os.path.join(self.stage_dir, "before-stage"), FileExists())
        self.assertThat(os.path.join(self.stage_dir, "after-stage"), FileExists())

        # Also assert that, while file2 was installed, it wasn't staged
        self.assertThat(
            os.path.join(
                self.parts_dir, "override-stage-do-nothing", "install", "file2"
            ),
            FileExists(),
        )
        self.assertThat(os.path.join(self.stage_dir, "file2"), Not(FileExists()))

    def test_override_prime(self):
        self.run_snapcraft("prime", "scriptlet-override-prime")

        installdir = os.path.join(
            self.parts_dir, "override-prime-scriptlet-test", "install"
        )

        # Assert that the before/after prime files aren't placed in the
        # installdir, although the file is.
        self.assertThat(os.path.join(installdir, "file"), FileExists())
        self.assertThat(os.path.join(installdir, "before-prime"), Not(FileExists()))
        self.assertThat(os.path.join(installdir, "after-prime"), Not(FileExists()))

        # Assert that the before/after prime files aren't placed in the
        # stagedir, although the file is.
        self.assertThat(os.path.join(self.stage_dir, "file"), FileExists())
        self.assertThat(os.path.join(self.stage_dir, "before-prime"), Not(FileExists()))
        self.assertThat(os.path.join(self.stage_dir, "after-prime"), Not(FileExists()))

        self.assertThat(os.path.join(self.prime_dir, "file"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "before-prime"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "after-prime"), FileExists())

        # Also assert that, while file2 was installed and staged, it wasn't
        # primed
        self.assertThat(
            os.path.join(
                self.parts_dir, "override-prime-do-nothing", "install", "file2"
            ),
            FileExists(),
        )
        self.assertThat(os.path.join(self.stage_dir, "file2"), FileExists())
        self.assertThat(os.path.join(self.prime_dir, "file2"), Not(FileExists()))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Any, Dict, Set

from snapcraft import project
from snapcraft.internal import repo
from snapcraft.internal.project_loader import grammar


class GlobalGrammarProcessor:
    """Process global properties that support grammar.

    Build packages example:
    >>> import snapcraft
    >>> from snapcraft import repo
    >>> processor = GlobalGrammarProcessor(
    ...    properties={'build-packages': [{'try': ['hello']}]},
    ...    project=snapcraft.project.Project())
    >>> processor.get_build_packages()
    {'hello'}
    """

    def __init__(self, *, properties: Dict[str, Any], project: project.Project) -> None:
        self._project = project

        self._build_package_grammar = properties.get("build-packages", [])
        self.__build_packages = set()  # type: Set[str]

    def get_build_packages(self) -> Set[str]:
        if not self.__build_packages:
            processor = grammar.GrammarProcessor(
                self._build_package_grammar,
                self._project,
                repo.Repo.build_package_is_valid,
            )
            self.__build_packages = set(processor.process())

        return self.__build_packages



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import doctest
import platform
import re

import pytest

import snapcraft
import snapcraft.internal.project_loader.grammar._to as to
from snapcraft.internal.project_loader import grammar


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(to))
    return tests


class TestToStatementGrammar:

    scenarios = [
        (
            "no target arch",
            {
                "to_arch": "to amd64",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": None,
                "expected_packages": ["foo"],
            },
        ),
        (
            "amd64 to armhf",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_packages": ["foo"],
            },
        ),
        (
            "amd64 to armhf, arch specified",
            {
                "to_arch": "to armhf",
                "body": ["foo:amd64"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_packages": ["foo:amd64"],
            },
        ),
        (
            "amd64 to i386",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "i386",
                "expected_packages": list(),
            },
        ),
        (
            "ignored else",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [["bar"]],
                "target_arch": "armhf",
                "expected_packages": ["foo"],
            },
        ),
        (
            "used else",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [["bar"]],
                "target_arch": "i386",
                "expected_packages": ["bar"],
            },
        ),
        (
            "used else, arch specified",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [["bar:amd64"]],
                "target_arch": "i386",
                "expected_packages": ["bar:amd64"],
            },
        ),
        (
            "third else ignored",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [["bar"], ["baz"]],
                "target_arch": "i386",
                "expected_packages": ["bar"],
            },
        ),
        (
            "third else followed",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [[{"to armhf": ["bar"]}], ["baz"]],
                "target_arch": "i386",
                "expected_packages": ["baz"],
            },
        ),
        (
            "nested armhf",
            {
                "to_arch": "to armhf",
                "body": [{"to armhf": ["foo"]}, {"to i386": ["bar"]}],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_packages": ["foo"],
            },
        ),
        (
            "nested i386",
            {
                "to_arch": "to i386",
                "body": [{"to armhf": ["foo"]}, {"to i386": ["bar"]}],
                "else_bodies": [],
                "target_arch": "i386",
                "expected_packages": ["bar"],
            },
        ),
        (
            "nested body ignored else",
            {
                "to_arch": "to armhf",
                "body": [{"to armhf": ["foo"]}, {"else": ["bar"]}],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_packages": ["foo"],
            },
        ),
        (
            "nested body used else",
            {
                "to_arch": "to i386",
                "body": [{"to armhf": ["foo"]}, {"else": ["bar"]}],
                "else_bodies": [],
                "target_arch": "i386",
                "expected_packages": ["bar"],
            },
        ),
        (
            "nested else ignored else",
            {
                "to_arch": "to i386",
                "body": ["foo"],
                "else_bodies": [[{"to armhf": ["bar"]}, {"else": ["baz"]}]],
                "target_arch": "armhf",
                "expected_packages": ["bar"],
            },
        ),
        (
            "nested else used else",
            {
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [[{"to armhf": ["bar"]}, {"else": ["baz"]}]],
                "target_arch": "i386",
                "expected_packages": ["baz"],
            },
        ),
    ]

    def test(
        self, monkeypatch, to_arch, body, else_bodies, target_arch, expected_packages
    ):
        monkeypatch.setattr(platform, "machine", lambda: "x86_64")
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))
        processor = grammar.GrammarProcessor(
            None, snapcraft.ProjectOptions(target_deb_arch=target_arch), lambda x: True
        )
        statement = to.ToStatement(to=to_arch, body=body, processor=processor)

        for else_body in else_bodies:
            statement.add_else(else_body)

        assert statement.process() == expected_packages


class TestToStatementInvalidGrammar:

    scenarios = [
        (
            "spaces in selectors",
            {
                "to_arch": "to armhf, ubuntu",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_exception": ".*not a valid 'to' clause.*spaces are not allowed in the "
                "selectors.*",
            },
        ),
        (
            "beginning with comma",
            {
                "to_arch": "to ,armhf",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_exception": ".*not a valid 'to' clause",
            },
        ),
        (
            "ending with comma",
            {
                "to_arch": "to armhf,",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_exception": ".*not a valid 'to' clause",
            },
        ),
        (
            "multiple commas",
            {
                "to_arch": "to armhf,,ubuntu",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_exception": ".*not a valid 'to' clause",
            },
        ),
        (
            "invalid selector format",
            {
                "to_arch": "to_arch",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_exception": ".*not a valid 'to' clause.*selectors are missing",
            },
        ),
        (
            "not even close",
            {
                "to_arch": "im-invalid",
                "body": ["foo"],
                "else_bodies": [],
                "target_arch": "armhf",
                "expected_exception": ".*not a valid 'to' clause",
            },
        ),
    ]

    def test(self, to_arch, body, else_bodies, target_arch, expected_exception):
        with pytest.raises(grammar.errors.ToStatementSyntaxError) as error:
            processor = grammar.GrammarProcessor(
                None,
                snapcraft.ProjectOptions(target_deb_arch=target_arch),
                lambda x: True,
            )
            statement = to.ToStatement(to=to_arch, body=body, processor=processor)

            for else_body in else_bodies:
                statement.add_else(else_body)

            statement.process()

        assert re.match(expected_exception, str(error.value))


def test_else_fail(monkeypatch):
    monkeypatch.setattr(platform, "machine", lambda: "x86_64")
    monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

    processor = grammar.GrammarProcessor(
        None, snapcraft.ProjectOptions(target_deb_arch="i386"), lambda x: True
    )
    statement = to.ToStatement(to="to armhf", body=["foo"], processor=processor)

    statement.add_else(None)

    with pytest.raises(grammar.errors.UnsatisfiedStatementError) as error:
        statement.process()

    assert str(error.value) == "Unable to satisfy 'to armhf', failure forced"



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

import fixtures
from testtools.matchers import Equals, FileExists

from snapcraft.internal.remote_build import InfoFile
from tests import unit


class TestInfoFile(unit.TestCase):
    def setUp(self):
        super().setUp()
        self._path = self.useFixture(fixtures.TempDir()).path

    def test_creation(self):
        path = os.path.join(self._path, "test.yaml")
        info = InfoFile(path)
        info.save()
        self.assertThat(path, FileExists())

    def test_save(self):
        path = os.path.join(self._path, "test.yaml")
        info = InfoFile(path)
        info.save(foo="bar")
        with open(path, "r") as f:
            data = f.readlines()
        self.assertThat(data, Equals(["foo: bar\n"]))

    def test_load(self):
        path = os.path.join(self._path, "test.yaml")
        with open(path, "w") as f:
            f.write("foo: 1\nbar:\n- a\n- b\n")
        info = InfoFile(path)
        info.load()
        self.assertThat(info["foo"], Equals(1))
        self.assertThat(info["bar"], Equals(["a", "b"]))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import platform
import re

import pytest

import snapcraft
import snapcraft.internal.project_loader.grammar._compound as compound
import snapcraft.internal.project_loader.grammar._on as on
import snapcraft.internal.project_loader.grammar._to as to
from snapcraft.internal.project_loader import grammar


class TestCompoundStatementGrammar:

    scenarios = [
        (
            "on amd64",
            {
                "on_arch": "on amd64",
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [],
                "host_arch": "x86_64",
                "expected_packages": ["foo"],
            },
        ),
        (
            "on i386",
            {
                "on_arch": "on amd64",
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [],
                "host_arch": "i686",
                "expected_packages": list(),
            },
        ),
        (
            "ignored else",
            {
                "on_arch": "on amd64",
                "to_arch": "to armhf",
                "body": ["foo"],
                "else_bodies": [["bar"]],
                "host_arch": "x86_64",
                "expected_packages": ["foo"],
            },
        ),
        (
            "used else",
            {
                "on_arch": "on amd64",
                "to_arch": "to i386",
                "body": ["foo"],
                "else_bodies": [["bar"]],
                "host_arch": "i686",
                "expected_packages": ["bar"],
            },
        ),
        (
            "third else ignored",
            {
                "on_arch": "on amd64",
                "to_arch": "to i386",
                "body": ["foo"],
                "else_bodies": [["bar"], ["baz"]],
                "host_arch": "i686",
                "expected_packages": ["bar"],
            },
        ),
        (
            "third else followed",
            {
                "on_arch": "on amd64",
                "to_arch": "to i386",
                "body": ["foo"],
                "else_bodies": [[{"on armhf": ["bar"]}], ["baz"]],
                "host_arch": "i686",
                "expected_packages": ["baz"],
            },
        ),
        (
            "nested amd64",
            {
                "on_arch": "on amd64",
                "to_arch": "to armhf",
                "body": [{"on amd64": ["foo"]}, {"on i386": ["bar"]}],
                "else_bodies": [],
                "host_arch": "x86_64",
                "expected_packages": ["foo"],
            },
        ),
        (
            "nested i386",
            {
                "on_arch": "on i386",
                "to_arch": "to armhf",
                "body": [{"on amd64": ["foo"]}, {"on i386": ["bar"]}],
                "else_bodies": [],
                "host_arch": "i686",
                "expected_packages": ["bar"],
            },
        ),
        (
            "nested body ignored else",
            {
                "on_arch": "on amd64",
                "to_arch": "to armhf",
                "body": [{"on amd64": ["foo"]}, {"else": ["bar"]}],
                "else_bodies": [],
                "host_arch": "x86_64",
                "expected_packages": ["foo"],
            },
        ),
        (
            "nested body used else",
            {
                "on_arch": "on i386",
                "to_arch": "to armhf",
                "body": [{"on amd64": ["foo"]}, {"else": ["bar"]}],
                "else_bodies": [],
                "host_arch": "i686",
                "expected_packages": ["bar"],
            },
        ),
        (
            "nested else ignored else",
            {
                "on_arch": "on armhf",
                "to_arch": "to i386",
                "body": ["foo"],
                "else_bodies": [[{"on amd64": ["bar"]}, {"else": ["baz"]}]],
                "host_arch": "x86_64",
                "expected_packages": ["bar"],
            },
        ),
        (
            "nested else used else",
            {
                "on_arch": "on armhf",
                "to_arch": "to i386",
                "body": ["foo"],
                "else_bodies": [[{"on amd64": ["bar"]}, {"else": ["baz"]}]],
                "host_arch": "i686",
                "expected_packages": ["baz"],
            },
        ),
        (
            "with hyphen",
            {
                "on_arch": "on other-arch",
                "to_arch": "to yet-another-arch",
                "body": ["foo"],
                "else_bodies": [],
                "host_arch": "x86_64",
                "expected_packages": list(),
            },
        ),
        (
            "multiple selectors",
            {
                "on_arch": "on amd64,i386",
                "to_arch": "to armhf,arm64",
                "body": ["foo"],
                "else_bodies": [],
                "host_arch": "x86_64",
                "expected_packages": list(),
            },
        ),
    ]

    def test(
        self,
        monkeypatch,
        on_arch,
        to_arch,
        body,
        else_bodies,
        host_arch,
        expected_packages,
    ):
        monkeypatch.setattr(platform, "machine", lambda: host_arch)
        monkeypatch.setattr(platform, "architecture", lambda: ("64bit", "ELF"))

        processor = grammar.GrammarProcessor(
            None, snapcraft.ProjectOptions(target_deb_arch="armhf"), lambda x: True
        )
        statements = [
            on.OnStatement(on=on_arch, body=None, processor=processor),
            to.ToStatement(to=to_arch, body=None, processor=processor),
        ]
        statement = compound.CompoundStatement(
            statements=statements, body=body, processor=processor
        )

        for else_body in else_bodies:
            statement.add_else(else_body)

        assert statement.process() == expected_packages


class TestCompoundStatementInvalidGrammar:

    scenarios = [
        (
            "spaces in on selectors",
            {
                "on_arch": "on amd64, ubuntu",
                "to_arch": "to i386",
                "body": ["foo"],
                "else_bodies": [],
                "expected_exception": grammar.errors.OnStatementSyntaxError,
                "expected_message": ".*not a valid 'on' clause.*spaces are not allowed in the "
                "selectors.*",
            },
        ),
        (
            "spaces in to selectors",
            {
                "on_arch": "on amd64,ubuntu",
                "to_arch": "to i386, armhf",
                "body": ["foo"],
                "else_bodies": [],
                "expected_exception": grammar.errors.ToStatementSyntaxError,
                "expected_message": ".*not a valid 'to' clause.*spaces are not allowed in the "
                "selectors.*",
            },
        ),
    ]

    def test(
        self, on_arch, to_arch, body, else_bodies, expected_exception, expected_message
    ):
        with pytest.raises(expected_exception) as error:
            processor = grammar.GrammarProcessor(
                None,
                snapcraft.ProjectOptions(target_deb_arch="armhf"),
                lambda x: "invalid" not in x,
            )
            statements = [
                on.OnStatement(on=on_arch, body=None, processor=processor),
                to.ToStatement(to=to_arch, body=None, processor=processor),
            ]
            statement = compound.CompoundStatement(
                statements=statements, body=body, processor=processor
            )

            for else_body in else_bodies:
                statement.add_else(else_body)

            statement.process()

        assert re.match(expected_message, str(error.value))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import distutils.util
import logging
import os
import shutil
import sys
import tempfile
import traceback
from textwrap import dedent
from typing import Dict

import click
from raven import Client as RavenClient
from raven.transport import RequestsHTTPTransport

import snapcraft
from snapcraft.config import CLIConfig as _CLIConfig
from snapcraft.internal import errors

from . import echo

# TODO:
# - annotate the part and lifecycle step in the message
# - add link to privacy policy
_MSG_TRACEBACK_PRINT = "Sorry, an error occurred in Snapcraft:"
_MSG_TRACEBACK_LOCATION = "You can find the traceback in file {!r}."
_MSG_SEND_TO_SENTRY_TRACEBACK_PROMPT = dedent(
    """\
    We would appreciate it if you anonymously reported this issue.
    No other data than the traceback and the version of snapcraft in use will be sent.
    Would you like to send this error data? (Yes/No/Always/View)"""
)
_MSG_SEND_TO_SENTRY_THANKS = "Thank you, sent."
_MSG_SILENT_REPORT = (
    "Sending an error report because SNAPCRAFT_ENABLE_SILENT_REPORT is set."
)
_MSG_ALWAYS_REPORT = dedent(
    """\
    Sending an error report because ALWAYS was selected in a past prompt.
    This behavior can be changed by changing the always_send entry in {}."""
)
_MSG_MANUALLY_REPORT = dedent(
    """\
    We would appreciate it if you created a bug report at
    https://launchpad.net/snapcraft/+filebug with the above text included."""
)

_YES_VALUES = ["yes", "y"]
_NO_VALUES = ["no", "n"]
_ALWAYS_VALUES = ["always", "a"]
_VIEW_VALUES = ["view", "v"]

TRACEBACK_MANAGED = os.path.join(tempfile.gettempdir(), "snapcraft_provider_traceback")
TRACEBACK_HOST = TRACEBACK_MANAGED + ".{}".format(os.getpid())


logger = logging.getLogger(__name__)


def _is_connected_to_tty() -> bool:
    # Used by inner instance, SNAPCRAFT_HAS_TTY is set by outer instance.
    if os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT") == "managed-host":
        return distutils.util.strtobool(os.getenv("SNAPCRAFT_HAS_TTY", "n")) == 1

    return sys.stdout.isatty()


def _is_reportable_error(exc_info) -> bool:
    # SnapcraftException has explicit `repotable` attribute.
    if isinstance(exc_info[1], errors.SnapcraftException):
        return exc_info[1].get_reportable()

    # Report non-snapcraft errors.
    if not issubclass(exc_info[0], errors.SnapcraftError) and not isinstance(
        exc_info[1], KeyboardInterrupt
    ):
        return True

    # Report SnapcraftReportableError errors.
    if issubclass(exc_info[0], errors.SnapcraftReportableError):
        return True

    return False


def _is_printable_traceback(exc_info, debug) -> bool:
    # Always print with debug.
    if debug:
        return True

    # If it is not a reportable error, then it is not a printable traceback.
    if not _is_reportable_error(exc_info):
        return False

    # Print if not connected to a tty.
    if not _is_connected_to_tty():
        return True

    # Print if not using snap.
    if not snapcraft.internal.common.is_snap():
        return True

    return False


def _handle_sentry_submission(exc_info) -> None:
    # Only attempt to submit reportable errors.
    if not _is_reportable_error(exc_info):
        return

    # Only attempt if running as snap (with Raven requirement).
    if not snapcraft.internal.common.is_snap():
        # Suggest manual reporting instead.
        click.echo(_MSG_MANUALLY_REPORT)
        return

    if _is_send_to_sentry(exc_info):
        try:
            _submit_trace(exc_info)
        except Exception as exc:
            # Failed to send - let user know and suggest manual reporting.
            echo.error(
                "Encountered an issue while trying to submit the report: {}".format(
                    str(exc)
                )
            )
            click.echo(_MSG_MANUALLY_REPORT)
        else:
            # Sent successfully.
            click.echo(_MSG_SEND_TO_SENTRY_THANKS)


def _print_snapcraft_exception_message(exception: errors.SnapcraftException):
    parts = [exception.get_brief()]

    resolution = exception.get_resolution()
    if resolution:
        parts.extend(["", "Recommended resolution:", resolution])

    details = exception.get_details()
    if details:
        parts.extend(["", "Detailed information:", details])

    docs_url = exception.get_docs_url()
    if docs_url:
        parts.extend(["", "For more information, check out:", docs_url])

    echo.error("\n".join(parts))


def _print_exception_message(exception: Exception) -> None:
    if isinstance(exception, errors.SnapcraftException):
        _print_snapcraft_exception_message(exception)
    else:
        echo.error(str(exception))


def _process_exception(exc_info, debug, trace_filepath):
    _print_exception_message(exc_info[1])

    if _is_printable_traceback(exc_info, debug):
        _print_trace_output(exc_info)

    _handle_sentry_submission(exc_info)

    if trace_filepath:
        with open(trace_filepath, "w") as tf:
            _print_trace_output(exc_info, tf)


def _process_inner_exception(exc_info, debug):
    # On inner host, always save trace output and process exception.
    _process_exception(exc_info, debug, TRACEBACK_MANAGED)


def _process_outer_exception(exc_info, debug):
    # Temporary file path to store trace in, if needed.
    trace_filepath = os.path.join(tempfile.mkdtemp(), "trace.txt")

    # On outer host, check if we have traceback generated from inner.
    if os.path.isfile(TRACEBACK_HOST):
        # If traceback file exists, it has already been processed by
        # the inside snapcraft (i.e. printing and sentry submission)
        # we just need to retrieve it for the user if it's reportable.
        if not _is_reportable_error(exc_info):
            return

        shutil.move(TRACEBACK_HOST, trace_filepath)
    else:
        # No traceback found, this must be captured and processed.
        # If reportable or debug is enabled, capture trace file.
        if _is_reportable_error(exc_info) or debug:
            # Only show this message for reportable errors.
            click.echo(_MSG_TRACEBACK_PRINT)

            _process_exception(exc_info, debug, trace_filepath)
        else:
            _process_exception(exc_info, debug, None)

            # No trace file, nothing left to do.
            return

    # This is a reportable error, let the user know where to find the trace.
    click.echo(_MSG_TRACEBACK_LOCATION.format(trace_filepath))


def _get_exception_exit_code(exception: Exception):
    if isinstance(exception, errors.SnapcraftException):
        return exception.get_exit_code()

    if isinstance(exception, errors.SnapcraftError):
        return exception.get_exit_code()

    # If non-snapcraft error, exit code defaults to 1.
    return 1


def exception_handler(  # noqa: C901
    exception_type, exception, exception_traceback, *, debug=False
) -> None:
    """Catch all Snapcraft exceptions unless debugging.

    This function is the global excepthook, properly handling uncaught
    exceptions and determine if they need to be reported.

    These are the rules of engagement:

        - Print exception message (always)

        - Capture trace in temporary file and print path, conditionally if:
          (1) Exception is SnapcraftReportableError or
          (2) Exception is non-Snapcraft error
          (3) debug is true

        - Automatically send bug report if:
          (1) SNAPCRAFT_ENABLE_SILENT_REPORT=y
          (2) "Always" send has been previously set.

        - Prompt to send bug report, conditionally if connected to TTY and:
          (1) Exception is SnapcraftReportableError or
          (2) Exception is non-Snapcraft error

        - Print trace, conditionally if:
          (1) debug is true, or
          (2) Reportable (see above) and either:
              (a) no TTY
              (b) not running as snap

        - Use exit code from snapcraft error (if available), otherwise 1.
    """
    exit_code = _get_exception_exit_code(exception)
    exc_info = (exception_type, exception, exception_traceback)

    if os.getenv("SNAPCRAFT_BUILD_ENVIRONMENT") == "managed-host":
        _process_inner_exception(exc_info, debug)
    else:
        _process_outer_exception(exc_info, debug)

    sys.exit(exit_code)


def _print_trace_output(exc_info, file=sys.stdout) -> None:
    # mypy does not like *exc_info with a kwarg that follows
    # snapcraft/cli/_errors.py:132: error: "print_exception" gets multiple values for keyword argument "file"
    traceback.print_exception(exc_info[0], exc_info[1], exc_info[2], file=file)


def _is_send_to_sentry(exc_info) -> bool:  # noqa: C901
    # Check to see if error reporting has been disabled
    if (
        distutils.util.strtobool(os.getenv("SNAPCRAFT_ENABLE_ERROR_REPORTING", "y"))
        == 0
    ):
        return False

    # Check the environment to see if we should allow for silent reporting
    if distutils.util.strtobool(os.getenv("SNAPCRAFT_ENABLE_SILENT_REPORT", "n")) == 1:
        click.echo(_MSG_SILENT_REPORT)
        return True

    # If ALWAYS has already been selected from before do not even bother to
    # prompt again.
    config_errors = None
    try:
        with _CLIConfig(read_only=True) as cli_config:
            if cli_config.get_sentry_send_always():
                click.echo(_MSG_ALWAYS_REPORT.format(cli_config.config_path))
                return True
    except errors.SnapcraftInvalidCLIConfigError as config_error:
        echo.warning(
            "There was an issue while trying to retrieve "
            "configuration data: {!s}".format(config_error)
        )
        config_errors = config_error

    # Either ALWAYS has not been selected in a previous run or the
    # configuration for where that value is stored cannot be read, so
    # resort to prompting (if connected to TTY).
    if not _is_connected_to_tty():
        return False

    while True:
        try:
            response = _prompt_sentry()
        except click.exceptions.Abort:
            # This was most likely triggered by a KeyboardInterrupt so
            # adding a new line makes things look nice.
            print()
            return False

        if response in _VIEW_VALUES:
            _print_trace_output(exc_info)
        elif response in _YES_VALUES:
            return True
        elif response in _NO_VALUES:
            return False
        elif response in _ALWAYS_VALUES:
            if config_errors is not None:
                echo.warning(
                    "Not saving choice to always send data to Sentry as "
                    "the configuration file is corrupted.\n"
                    "Please edit and fix or alternatively remove the "
                    "configuration file {!r} from disk.".format(
                        config_errors.config_file
                    )
                )  # type: ignore
            else:
                click.echo("Saving choice to always send data to Sentry")
                with _CLIConfig() as cli_config:
                    cli_config.set_sentry_send_always(True)
            return True
        else:
            echo.warning("Could not determine choice, assuming no.")
            return False


def _prompt_sentry():
    msg = _MSG_SEND_TO_SENTRY_TRACEBACK_PROMPT
    all_valid = _YES_VALUES + _NO_VALUES + _ALWAYS_VALUES + _VIEW_VALUES

    def validate(value):
        if value.lower() in all_valid:
            return value
        raise click.BadParameter("Please choose a valid answer.")

    return echo.prompt(msg, default="no", value_proc=validate).lower()


def _submit_trace(exc_info):
    kwargs: Dict[str, str] = dict()
    if "+git" not in snapcraft.__version__:
        kwargs["release"] = snapcraft.__version__

    client = RavenClient(
        "https://b0fef3e0ced2443c92143ae0d038b0a4:"
        "b7c67d7fa4ee46caae12b29a80594c54@sentry.io/277754",
        transport=RequestsHTTPTransport,
        # Should Raven automatically log frame stacks (including locals)
        # for all calls as it would for exceptions.
        auto_log_stacks=False,
        # Set a name to not send the real hostname.
        name="snapcraft",
        # Removes all stacktrace context variables. This will cripple the
        # functionality of Sentry, as you’ll only get raw tracebacks,
        # but it will ensure no local scoped information is available to the
        # server.
        processors=(
            "raven.processors.RemoveStackLocalsProcessor",
            "raven.processors.SanitizePasswordsProcessor",
        ),
        **kwargs
    )
    client.captureException(exc_info=exc_info)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import hashlib
import json
import os

from snapcraft.storeapi.errors import StoreMetadataError


def _media_hash(media_file):
    sha = hashlib.sha256(media_file.read())
    # rewind file before returning
    media_file.seek(0)
    return sha.hexdigest()


class StoreMetadataHandler:
    def __init__(self, *, request_method, snap_id: str, snap_name: str):
        super().__init__()
        self._request = request_method
        self.snap_id = snap_id
        self.snap_name = snap_name

    def upload(self, metadata, force):
        """Upload the metadata to Dashboard."""
        url = f"/dev/api/snaps/{self.snap_id}/metadata"
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }
        method = "PUT" if force else "POST"
        response = self._request(
            method, url, data=json.dumps(metadata), headers=headers
        )

        if not response.ok:
            raise StoreMetadataError(self.snap_name, response, metadata)

    def _current_binary_metadata(self):
        """Get current icons and screenshots as set in the store."""
        url = f"/dev/api/snaps/{self.snap_id}/binary-metadata"
        headers = {"Accept": "application/json"}
        # get current binary metadata information
        response = self._request("GET", url, headers=headers)
        if not response.ok:
            raise StoreMetadataError(self.snap_name, response, {})

        binary_metadata = response.json()
        # current icons and screenshots
        icons = [media for media in binary_metadata if media.get("type") == "icon"]
        screenshots = [
            media for media in binary_metadata if media.get("type") == "screenshot"
        ]
        return icons, screenshots

    def _build_binary_request_data(self, metadata):
        """Return calculated data and files for the binary request."""
        data = files = None

        icons, screenshots = self._current_binary_metadata()
        current_icon = icons[0] if icons else None

        # only icon support atm
        icon = metadata.get("icon")
        # keep original screenshots
        updated_info = screenshots

        if current_icon is None and icon is None:
            # icon unchanged, nothing to upload
            return data, files

        if icon:
            icon_hash = _media_hash(icon)
            if current_icon is None or current_icon.get("hash") != icon_hash:
                upload_icon = {
                    "type": "icon",
                    "hash": icon_hash,
                    "key": "icon",
                    "filename": icon.name,
                }
                updated_info.append(upload_icon)
                files = {"icon": icon}
            else:
                # icon unchanged, nothing to upload
                return data, files

        if not files:
            # API requires a multipart request, but we have no files to upload
            # https://github.com/requests/requests/issues/1081
            files = {"info": ("", json.dumps(updated_info))}
        else:
            data = {"info": json.dumps(updated_info)}

        return data, files

    def upload_binary(self, metadata, force):
        """Upload the binary metadata to SCA."""
        data, files = self._build_binary_request_data(metadata)
        if data is None and files is None:
            # nothing to update
            return

        url = f"/dev/api/snaps/{self.snap_id}/binary-metadata"
        headers = {
            "Accept": "application/json",
        }
        method = "PUT" if force else "POST"
        response = self._request(method, url, data=data, files=files, headers=headers)
        if not response.ok:
            icon = metadata.get("icon")
            icon_name = os.path.basename(icon.name) if icon else None
            raise StoreMetadataError(self.snap_name, response, {"icon": icon_name})



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015, 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import pathlib
import subprocess
from unittest import mock

from testtools.matchers import Equals

import snapcraft
from snapcraft.plugins.v1._ros import rosdep
from tests import unit


class RosdepTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.project = snapcraft.ProjectOptions()

        self.rosdep = rosdep.Rosdep(
            ros_distro="kinetic",
            ros_version="1",
            ros_package_path="package_path",
            rosdep_path="rosdep_path",
            ubuntu_distro="xenial",
            base="core",
            target_arch=self.project._get_stage_packages_target_arch(),
        )

        patcher = mock.patch("snapcraft.repo.Ubuntu")
        self.ubuntu_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.check_output")
        self.check_output_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_setup(self):
        # Return something other than a Mock to ease later assertions
        self.check_output_mock.return_value = b""

        self.rosdep.setup()

        # Verify that only rosdep was installed (no other .debs)
        self.assertThat(
            self.ubuntu_mock.fetch_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=self.rosdep._rosdep_stage_packages_path,
                        package_names=["python-rosdep"],
                        base="core",
                        target_arch=self.project._get_stage_packages_target_arch(),
                    )
                ]
            ),
        )
        self.assertThat(
            self.ubuntu_mock.unpack_stage_packages.mock_calls,
            Equals(
                [
                    mock.call(
                        stage_packages_path=self.rosdep._rosdep_stage_packages_path,
                        install_path=pathlib.Path("rosdep_path/install"),
                    )
                ]
            ),
        )

        # Verify that rosdep was initialized and updated
        self.assertThat(self.check_output_mock.call_count, Equals(2))
        self.check_output_mock.assert_has_calls(
            [
                mock.call(["rosdep", "init"], env=mock.ANY),
                mock.call(["rosdep", "update", "--include-eol-distros"], env=mock.ANY),
            ]
        )

    def test_setup_can_run_multiple_times(self):
        self.rosdep.setup()

        # Make sure running setup() again doesn't have problems with the old
        # environment
        # An exception will be raised if setup can't be called twice.
        self.rosdep.setup()

    def test_setup_initialization_failure(self):
        def run(args, **kwargs):
            if args == ["rosdep", "init"]:
                raise subprocess.CalledProcessError(1, "foo", b"bar")

        self.check_output_mock.side_effect = run

        raised = self.assertRaises(rosdep.RosdepInitializationError, self.rosdep.setup)

        self.assertThat(
            str(raised),
            Equals(
                "Failed to initialize rosdep: Error initializing rosdep database:\nbar"
            ),
        )

    def test_setup_update_failure(self):
        def run(args, **kwargs):
            if args == ["rosdep", "update", "--include-eol-distros"]:
                raise subprocess.CalledProcessError(1, "foo", b"bar")

            return mock.DEFAULT

        self.check_output_mock.side_effect = run

        raised = self.assertRaises(rosdep.RosdepInitializationError, self.rosdep.setup)

        self.assertThat(
            str(raised),
            Equals("Failed to initialize rosdep: Error updating rosdep database:\nbar"),
        )

    def test_get_dependencies(self):
        self.check_output_mock.return_value = b"foo\nbar\nbaz"

        self.assertThat(
            self.rosdep.get_dependencies("foo"), Equals({"foo", "bar", "baz"})
        )

        self.check_output_mock.assert_called_with(
            ["rosdep", "keys", "foo"], env=mock.ANY
        )

    def test_get_dependencies_no_dependencies(self):
        self.check_output_mock.return_value = b""

        self.assertThat(self.rosdep.get_dependencies("foo"), Equals(set()))

    def test_get_dependencies_invalid_package(self):
        self.check_output_mock.side_effect = subprocess.CalledProcessError(1, "foo")

        raised = self.assertRaises(
            rosdep.RosdepPackageNotFoundError, self.rosdep.get_dependencies, "bar"
        )

        self.assertThat(str(raised), Equals("rosdep cannot find Catkin package 'bar'"))

    def test_get_dependencies_entire_workspace(self):
        self.check_output_mock.return_value = b"foo\nbar\nbaz"

        self.assertThat(self.rosdep.get_dependencies(), Equals({"foo", "bar", "baz"}))

        self.check_output_mock.assert_called_with(
            ["rosdep", "keys", "-a", "-i"], env=mock.ANY
        )

    def test_resolve_dependency(self):
        self.check_output_mock.return_value = b"#apt\nmylib-dev"

        self.assertThat(
            self.rosdep.resolve_dependency("foo"), Equals({"apt": {"mylib-dev"}})
        )

        self.check_output_mock.assert_called_with(
            [
                "rosdep",
                "resolve",
                "foo",
                "--rosdistro",
                "kinetic",
                "--os",
                "ubuntu:xenial",
            ],
            env=mock.ANY,
        )

    def test_resolve_invalid_dependency(self):
        self.check_output_mock.side_effect = subprocess.CalledProcessError(1, "foo")

        raised = self.assertRaises(
            rosdep.RosdepDependencyNotResolvedError,
            self.rosdep.resolve_dependency,
            "bar",
        )

        self.assertThat(
            str(raised), Equals("rosdep cannot resolve 'bar' into a valid dependency")
        )

    def test_resolve_unexpected_dependency(self):
        # Note the lack of dependency type here
        self.check_output_mock.return_value = b"mylib-dev"

        raised = self.assertRaises(
            rosdep.RosdepUnexpectedResultError, self.rosdep.resolve_dependency, "bar"
        )

        self.assertThat(
            str(raised),
            Equals(
                "Received unexpected result from rosdep when trying to "
                "resolve 'bar':\nmylib-dev"
            ),
        )

    def test_resolve_no_dependency(self):
        self.check_output_mock.return_value = b"#apt"

        self.assertThat(self.rosdep.resolve_dependency("bar"), Equals({"apt": set()}))

    def test_resolve_multiple_dependencies(self):
        self.check_output_mock.return_value = b"#apt\nlib1 lib2"

        self.assertThat(
            self.rosdep.resolve_dependency("foo"), Equals({"apt": {"lib1", "lib2"}})
        )

    def test_resolve_multiple_dependency_types(self):
        self.check_output_mock.return_value = b"#apt\nlib1\n\n#pip\nlib2"

        self.assertThat(
            self.rosdep.resolve_dependency("foo"),
            Equals({"apt": {"lib1"}, "pip": {"lib2"}}),
        )

    def test_run(self):
        rosdep = self.rosdep
        rosdep._run(["qux"])

        class check_env:
            def __eq__(self, env):
                rosdep_sources_path = rosdep._rosdep_sources_path
                return (
                    env["PATH"]
                    == os.path.join(rosdep._rosdep_install_path, "usr", "bin")
                    and env["PYTHONPATH"]
                    == os.path.join(
                        rosdep._rosdep_install_path,
                        "usr",
                        "lib",
                        "python2.7",
                        "dist-packages",
                    )
                    and env["ROSDEP_SOURCE_PATH"] == rosdep_sources_path
                    and env["ROS_HOME"] == rosdep._rosdep_cache_path
                    and env["ROS_PACKAGE_PATH"] == rosdep._ros_package_path
                    and env["ROS_VERSION"] == rosdep._ros_version
                )

        self.check_output_mock.assert_called_with(mock.ANY, env=check_env())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import click
import fixtures
from testtools.matchers import Equals

import snapcraft.cli._options as options
from tests import unit


class TestProviderOptions:
    scenarios = [
        ("host empty", dict(provider="host", kwargs=dict(), flags=dict())),
        (
            "host http proxy",
            dict(
                provider="host",
                kwargs=dict(http_proxy="1.1.1.1"),
                flags=dict(http_proxy="1.1.1.1"),
            ),
        ),
        (
            "host https proxy",
            dict(
                provider="host",
                kwargs=dict(https_proxy="1.1.1.1"),
                flags=dict(https_proxy="1.1.1.1"),
            ),
        ),
        (
            "host build info",
            dict(
                provider="host",
                kwargs=dict(enable_manifest=True),
                flags=dict(SNAPCRAFT_BUILD_INFO=True),
            ),
        ),
        (
            "host build info off",
            dict(provider="host", kwargs=dict(enable_manifest=False), flags=dict()),
        ),
        (
            "host image info",
            dict(
                provider="host",
                kwargs=dict(manifest_image_information="{}"),
                flags=dict(SNAPCRAFT_IMAGE_INFO="{}"),
            ),
        ),
        (
            "host all",
            dict(
                provider="host",
                kwargs=dict(http_proxy="1.1.1.1", https_proxy="1.1.1.1"),
                flags=dict(http_proxy="1.1.1.1", https_proxy="1.1.1.1"),
            ),
        ),
        ("lxd empty", dict(provider="lxd", kwargs=dict(), flags=dict())),
        (
            "lxd http proxy",
            dict(
                provider="lxd",
                kwargs=dict(http_proxy="1.1.1.1"),
                flags=dict(http_proxy="1.1.1.1"),
            ),
        ),
        (
            "lxd https proxy",
            dict(
                provider="lxd",
                kwargs=dict(https_proxy="1.1.1.1"),
                flags=dict(https_proxy="1.1.1.1"),
            ),
        ),
        (
            "lxd build info",
            dict(
                provider="lxd",
                kwargs=dict(enable_manifest=True),
                flags=dict(SNAPCRAFT_BUILD_INFO=True),
            ),
        ),
        (
            "lxd image info",
            dict(
                provider="lxd",
                kwargs=dict(manifest_image_information="{}"),
                flags=dict(SNAPCRAFT_IMAGE_INFO="{}"),
            ),
        ),
        (
            "lxd all",
            dict(
                provider="lxd",
                kwargs=dict(
                    http_proxy="1.1.1.1",
                    https_proxy="1.1.1.1",
                    enable_manifest=True,
                    manifest_image_information="{}",
                ),
                flags=dict(
                    http_proxy="1.1.1.1",
                    https_proxy="1.1.1.1",
                    SNAPCRAFT_BUILD_INFO=True,
                    SNAPCRAFT_IMAGE_INFO="{}",
                ),
            ),
        ),
        (
            "managed-host empty",
            dict(provider="managed-host", kwargs=dict(), flags=dict()),
        ),
        (
            "managed-host http proxy",
            dict(
                provider="managed-host",
                kwargs=dict(http_proxy="1.1.1.1"),
                flags=dict(http_proxy="1.1.1.1"),
            ),
        ),
        (
            "managed-host https proxy",
            dict(
                provider="managed-host",
                kwargs=dict(https_proxy="1.1.1.1"),
                flags=dict(https_proxy="1.1.1.1"),
            ),
        ),
        (
            "managed-host build info",
            dict(
                provider="managed-host",
                kwargs=dict(enable_manifest=True),
                flags=dict(SNAPCRAFT_BUILD_INFO=True),
            ),
        ),
        (
            "managed-host image info",
            dict(
                provider="managed-host",
                kwargs=dict(manifest_image_information="{}"),
                flags=dict(SNAPCRAFT_IMAGE_INFO="{}"),
            ),
        ),
        (
            "managed-host all",
            dict(
                provider="managed-host",
                kwargs=dict(
                    http_proxy="1.1.1.1",
                    https_proxy="1.1.1.1",
                    enable_manifest=True,
                    manifest_image_information="{}",
                ),
                flags=dict(
                    http_proxy="1.1.1.1",
                    https_proxy="1.1.1.1",
                    SNAPCRAFT_BUILD_INFO=True,
                    SNAPCRAFT_IMAGE_INFO="{}",
                ),
            ),
        ),
        ("multipass empty", dict(provider="multipass", kwargs=dict(), flags=dict())),
        (
            "multipass http proxy",
            dict(
                provider="multipass",
                kwargs=dict(http_proxy="1.1.1.1"),
                flags=dict(http_proxy="1.1.1.1"),
            ),
        ),
        (
            "multipass https proxy",
            dict(
                provider="multipass",
                kwargs=dict(https_proxy="1.1.1.1"),
                flags=dict(https_proxy="1.1.1.1"),
            ),
        ),
        (
            "multipass build info",
            dict(
                provider="multipass",
                kwargs=dict(enable_manifest=True),
                flags=dict(SNAPCRAFT_BUILD_INFO=True),
            ),
        ),
        (
            "multipass image info",
            dict(
                provider="multipass",
                kwargs=dict(manifest_image_information="{}"),
                flags=dict(SNAPCRAFT_IMAGE_INFO="{}"),
            ),
        ),
        (
            "multipass all",
            dict(
                provider="multipass",
                kwargs=dict(
                    http_proxy="1.1.1.1",
                    https_proxy="1.1.1.1",
                    enable_manifest=True,
                    manifest_image_information="{}",
                ),
                flags=dict(
                    http_proxy="1.1.1.1",
                    https_proxy="1.1.1.1",
                    SNAPCRAFT_BUILD_INFO=True,
                    SNAPCRAFT_IMAGE_INFO="{}",
                ),
            ),
        ),
    ]

    def test(self, provider, kwargs, flags):
        assert options.get_build_provider_flags(provider, **kwargs) == flags


class TestInvalidBuildProviderFlags(unit.TestCase):
    def setUp(self):
        super().setUp()
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", None)
        )

    def test_ignored(self):
        kwargs = dict(notrelevant="foo")

        flags = options.get_build_provider_flags("lxd", **kwargs)

        self.assertThat(flags, Equals(dict()))

    def test_host_missing_destructive_mode(self):
        kwargs = dict(http_proxy="192.168.1.1")

        self.assertRaisesRegex(
            click.exceptions.BadArgumentUsage,
            "^--provider=host requires --destructive-mode to acknowledge side effects$",
            options._sanity_check_build_provider_flags,
            "host",
            **kwargs,
        )

    def test_build_environment_mismatch(self):
        kwargs = dict(http_proxy="192.168.1.1")

        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "managed-host")
        )

        self.assertRaisesRegex(
            click.exceptions.BadArgumentUsage,
            "^mismatch between --provider=lxd and SNAPCRAFT_BUILD_ENVIRONMENT=managed-host$",
            options._sanity_check_build_provider_flags,
            "lxd",
            **kwargs,
        )

    def test_wrong_provider_option_error(self):
        kwargs = dict(http_proxy="192.168.1.1")
        fake_args = ["--destructive-mode"]

        with mock.patch("sys.argv", fake_args):
            self.assertRaisesRegex(
                click.exceptions.BadArgumentUsage,
                "^--destructive-mode cannot be used with build provider 'lxd'$",
                options._sanity_check_build_provider_flags,
                "lxd",
                **kwargs,
            )

    def test_unknown_provider(self):
        kwargs = dict(http_proxy="192.168.1.1")
        fake_args = ["--http-proxy", "192.168.1.1"]

        with mock.patch("sys.argv", fake_args):
            self.assertRaisesRegex(
                click.exceptions.BadArgumentUsage,
                "^--http-proxy cannot be used with build provider 'invalid-provider'$",
                options._sanity_check_build_provider_flags,
                "invalid-provider",
                **kwargs,
            )


class TestSudo(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_BUILD_ENVIRONMENT", "host")
        )
        self.useFixture(fixtures.EnvironmentVariable("SUDO_USER", "testuser"))
        self.fake_euid = self.useFixture(
            fixtures.MockPatch("os.geteuid", return_value=0)
        ).mock

    @mock.patch("snapcraft.cli._options.warning")
    def test_click_warn_sudo(self, warning_mock):
        options._sanity_check_build_provider_flags("host")
        warning_mock.assert_called_once_with(
            "Running with 'sudo' may cause permission errors and is discouraged. Use 'sudo' when cleaning."
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess

from testtools.matchers import Contains, Equals

from tests import integration


class AfterTestCase(integration.TestCase):
    def test_stage_dependencies(self):
        self.run_snapcraft("stage", "dependencies")

        self.assertTrue(os.access(os.path.join(self.stage_dir, "bin", "p3"), os.X_OK))

    def test_build_with_circular_dependencies(self):
        self.copy_project_to_cwd("dependencies")

        with open("snapcraft.yaml", "r") as snapcraft_yaml:
            snapcraft_yaml_contents = snapcraft_yaml.read()
        with open("snapcraft.yaml", "w") as snapcraft_yaml:
            snapcraft_yaml.write(
                snapcraft_yaml_contents.replace("p1:", "p1:\n    after: [p3]")
            )

        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "build"
        )

        self.assertThat(exception.returncode, Equals(2))
        expected = (
            "Issue detected while analyzing snapcraft.yaml: "
            "circular dependency chain found in parts definition\n"
        )
        self.assertThat(exception.output, Contains(expected))

    def test_build_with_missing_dependencies(self):
        self.copy_project_to_cwd("dependencies")

        with open("snapcraft.yaml", "r") as snapcraft_yaml:
            snapcraft_yaml_contents = snapcraft_yaml.read()
        wrong_contents = snapcraft_yaml_contents.replace("    after: [p1]\n", "")
        wrong_contents = wrong_contents.replace("    after: [p2]\n", "")
        with open("snapcraft.yaml", "w") as snapcraft_yaml:
            snapcraft_yaml.write(wrong_contents)

        exception = self.assertRaises(
            subprocess.CalledProcessError, self.run_snapcraft, "build"
        )

        self.assertThat(exception.returncode, Equals(2))

    def test_pull_with_tree_of_dependencies(self):
        self.run_snapcraft("pull", os.path.join("circular-dependencies", "tree"))

    def test_pull_with_circular_dependencies(self):
        self.assertRaises(
            subprocess.CalledProcessError,
            self.run_snapcraft,
            "pull",
            os.path.join("circular-dependencies", "circle"),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from snapcraft.internal import elf, pluginhandler
from snapcraft.internal.project_loader import grammar_processing
from snapcraft.project import Project, _schema


def load_part(
    part_name,
    plugin_name=None,
    part_properties=None,
    project=None,
    stage_packages_repo=None,
    snap_name="test-snap",
    base="core18",
    build_base=None,
    confinement="strict",
    snap_type="app",
):
    if not plugin_name:
        plugin_name = "nil"
    properties = {"plugin": plugin_name}
    if part_properties:
        properties.update(part_properties)
    if "build-environment" not in properties:
        properties["build-environment"] = list()
    if not project:
        project = Project()

    project._snap_meta.name = snap_name
    project._snap_meta.version = "1.0"
    project._snap_meta.grade = "devel"
    project._snap_meta.type = snap_type
    project._snap_meta.confinement = confinement
    project._snap_meta.base = base
    if build_base is not None:
        project._snap_meta.build_base = build_base

    validator = _schema.Validator()
    schema = validator.part_schema
    definitions_schema = validator.definitions_schema
    plugin = pluginhandler.load_plugin(
        part_name=part_name,
        plugin_name=plugin_name,
        properties=properties,
        project=project,
        part_schema=schema,
        definitions_schema=definitions_schema,
    )

    if not stage_packages_repo:
        stage_packages_repo = mock.Mock()
    grammar_processor = grammar_processing.PartGrammarProcessor(
        plugin=plugin, properties=properties, project=project, repo=stage_packages_repo
    )

    return pluginhandler.PluginHandler(
        plugin=plugin,
        part_properties=properties,
        project=project,
        part_schema=schema,
        definitions_schema=definitions_schema,
        grammar_processor=grammar_processor,
        stage_packages_repo=stage_packages_repo,
        snap_base_path="/snap/fake-name/current",
        soname_cache=elf.SonameCache(),
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent
from unittest import mock

import pytest

from snapcraft.storeapi.v2 import validation_sets
from snapcraft.storeapi import StoreClient


@pytest.fixture
def fake_dashboard_get_validation_sets():
    patched_get_validation_sets = mock.patch.object(
        StoreClient,
        "get_validation_sets",
        return_value=validation_sets.ValidationSets.unmarshal({"assertions": []}),
    )
    yield patched_get_validation_sets.start()
    patched_get_validation_sets.stop()


combinations = [
    {
        "name": None,
        "sequence": None,
        "output": "No validation sets found for this account.",
    },
    {
        "name": "foo",
        "sequence": None,
        "output": "No validation sets found for the requested name or sequence.",
    },
    {
        "name": "foo",
        "sequence": "all",
        "output": "No validation sets found for the requested name or sequence.",
    },
    {
        "name": None,
        "sequence": "all",
        "output": "No validation sets found for the requested name or sequence.",
    },
]


@pytest.mark.parametrize("combo,", combinations)
def test_no_sets(click_run, fake_dashboard_get_validation_sets, combo):
    cmd = ["list-validation-sets"]
    if combo["name"] is not None:
        cmd.extend(["--name", combo["name"]])
    if combo["sequence"] is not None:
        cmd.extend(["--sequence", combo["sequence"]])

    result = click_run(cmd)

    assert result.exit_code == 0
    assert result.output.strip() == combo["output"]
    fake_dashboard_get_validation_sets.assert_called_once_with(
        name=combo["name"], sequence=combo["sequence"]
    )


def test_list_validation_sets(click_run, fake_dashboard_get_validation_sets):
    fake_dashboard_get_validation_sets.return_value = validation_sets.ValidationSets.unmarshal(
        {
            "assertions": [
                {
                    "headers": {
                        "account-id": "AccountIDXXXOfTheRequestingUserX",
                        "authority-id": "AccountIDXXXOfTheRequestingUserX",
                        "name": "certification-x1",
                        "revision": "222",
                        "sequence": "9",
                        "series": "16",
                        "sign-key-sha3-384": "XSignXKeyXHashXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                        "snaps": [
                            {
                                "id": "XXSnapIDForXSnapName1XXXXXXXXXXX",
                                "name": "snap-name-1",
                                "presence": "optional",
                            },
                            {
                                "id": "XXSnapIDForXSnapName2XXXXXXXXXXX",
                                "name": "snap-name-2",
                            },
                        ],
                        "timestamp": "2020-10-29T16:36:56Z",
                        "type": "validation-set",
                    }
                },
                {
                    "headers": {
                        "account-id": "AccountIDXXXOfTheRequestingUserX",
                        "authority-id": "AccountIDXXXOfTheRequestingUserX",
                        "name": "acme-qa",
                        "revision": "2",
                        "sequence": "2",
                        "series": "16",
                        "sign-key-sha3-384": "XSignXKeyXHashXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                        "snaps": [
                            {
                                "id": "XXSnapIDForXSnapName3XXXXXXXXXXX",
                                "name": "snap-name-3",
                                "presence": "required",
                                "revision": "2",
                            }
                        ],
                        "timestamp": "2020-10-29T16:36:56Z",
                        "type": "validation-set",
                    }
                },
            ]
        }
    )

    cmd = ["list-validation-sets"]

    result = click_run(cmd)

    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        Account-ID                        Name              Sequence    Revision    When
        AccountIDXXXOfTheRequestingUserX  certification-x1  9           222         2020-10-29
        AccountIDXXXOfTheRequestingUserX  acme-qa           2           2           2020-10-29
        """
    )
    fake_dashboard_get_validation_sets.assert_called_once_with(
        name=None, sequence=None,
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
from textwrap import dedent

from testtools.matchers import Equals

from snapcraft.internal import project_loader
from snapcraft.project import Project
from tests import fixture_setup

from . import LoadPartBaseTest, ProjectLoaderBaseTest


def get_project_config(snapcraft_yaml_content):
    snapcraft_yaml_path = pathlib.Path("Snapcraft.yaml")
    with snapcraft_yaml_path.open("w") as snapcraft_yaml_file:
        print(snapcraft_yaml_content, file=snapcraft_yaml_file)

    project = Project(snapcraft_yaml_file_path=snapcraft_yaml_path.as_posix())
    return project_loader.load_config(project)


class TestParts(ProjectLoaderBaseTest):
    def make_snapcraft_project(self, parts):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(self.path)
        snapcraft_yaml.update_part("part1", dict(plugin="nil"))
        for part_name, part in parts:
            snapcraft_yaml.update_part(part_name, part)
        self.useFixture(snapcraft_yaml)

        project = Project(
            snapcraft_yaml_file_path=snapcraft_yaml.snapcraft_yaml_file_path
        )
        return project_loader.load_config(project)

    def test_get_parts_none(self):
        project_config = self.make_snapcraft_project([("part1", dict(plugin="nil"))])
        self.assertThat(project_config.parts.get_part("not-a-part"), Equals(None))

    def test_after_inexistent_part(self):
        raised = self.assertRaises(
            project_loader.errors.SnapcraftAfterPartMissingError,
            self.make_snapcraft_project,
            [("part1", dict(plugin="nil", after=["inexistent-part"]))],
        )

        self.assertThat(raised.part_name, Equals("part1"))
        self.assertThat(raised.after_part_name, Equals("inexistent-part"))


class TestPartOrder:

    scenarios = [
        (
            "part1 then part2",
            {
                "contents": dedent(
                    """\
                name: test
                base: core18
                version: "1"
                summary: test
                description: test
                confinement: strict

                parts:
                  part1:
                    plugin: nil
                  part2:
                    plugin: nil
                """
                ),
                "expected_order": ["part1", "part2"],
            },
        ),
        (
            "part2 then part1",
            {
                "contents": dedent(
                    """\
                name: test
                base: core18
                version: "1"
                summary: test
                description: test
                confinement: strict

                parts:
                  part2:
                    plugin: nil
                  part1:
                    plugin: nil
                """
                ),
                "expected_order": ["part1", "part2"],
            },
        ),
        (
            "single after",
            {
                "contents": dedent(
                    """\
                name: test
                base: core18
                version: "1"
                summary: test
                description: test
                confinement: strict

                parts:
                  part2:
                    plugin: nil
                    after: [part3]
                  part1:
                    plugin: nil
                  part3:
                    plugin: nil
                """
                ),
                "expected_order": ["part1", "part3", "part2"],
            },
        ),
        (
            "multiple after",
            {
                "contents": dedent(
                    """\
                name: test
                base: core18
                version: "1"
                summary: test
                description: test
                confinement: strict

                parts:
                  part2:
                    plugin: nil
                    after: [part3]
                  part1:
                    plugin: nil
                    after: [part3]
                  part3:
                    plugin: nil
                """
                ),
                "expected_order": ["part3", "part1", "part2"],
            },
        ),
    ]

    def test_part_order_consistency(self, tmp_work_path, contents, expected_order):
        """Test that parts are always processed in the same order."""

        project_config = get_project_config(contents)

        assert len(project_config.all_parts) == len(expected_order)

        for part, expected_name in zip(project_config.all_parts, expected_order):
            assert part.name == expected_name


class PluginLoadTest(LoadPartBaseTest):
    def test_plugin_loading(self):
        self.make_snapcraft_project(
            dedent(
                """\
            name: test
            base: core18
            version: "1"
            summary: test
            description: test
            confinement: strict
            grade: stable

            parts:
              part1:
                plugin: nil
                stage-packages: [fswebcam]
            """
            )
        )

        self.mock_load_part.assert_called_with(
            "part1",
            "nil",
            {"stage-packages": ["fswebcam"], "plugin": "nil", "stage": [], "prime": []},
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from snapcraft.storeapi.v2 import releases


@pytest.mark.parametrize(
    "branch", ((None, None), ("test-branch", "2020-02-11T17:51:40.891996Z"))
)
@pytest.mark.parametrize("revision", (42, None))
def test_release(branch, revision):
    payload = {
        "architecture": "amd64",
        "branch": branch[0],
        "channel": "latest/stable",
        "expiration-date": branch[1],
        "revision": revision,
        "risk": "stable",
        "track": "latest",
        "when": "2020-02-11T17:51:40.891996Z",
    }

    r = releases.Release.unmarshal(payload)

    assert repr(r) == "<Release: 'latest/stable'>"
    assert r.architecture == payload["architecture"]
    assert r.branch == payload["branch"]
    assert r.channel == payload["channel"]
    assert r.expiration_date == payload["expiration-date"]
    assert r.revision == payload["revision"]
    assert r.risk == payload["risk"]
    assert r.track == payload["track"]
    assert r.when == payload["when"]
    assert r.marshal() == payload


@pytest.mark.parametrize("architectures", (["amd64"], ["amd64", "arm64"]))
@pytest.mark.parametrize("build_url", (None, "https://launchpad.net/foo"))
@pytest.mark.parametrize("base", (None, "core20"))
def test_revision(architectures, build_url, base):
    payload = {
        "architectures": architectures,
        "build_url": build_url,
        "confinement": "strict",
        "created_at": "2020-02-11T17:51:40.891996Z",
        "grade": "stable",
        "revision": 42,
        "sha3-384": "a9060ef4872ccacbfa440617a76fcd84967896b28d0d1eb7571f00a1098d766e7e93353b084ba6ad841d7b14b95ede48",
        "size": 20,
        "status": "Published",
        "version": "1.0",
    }
    if base is not None:
        payload["base"] = base

    r = releases.Revision.unmarshal(payload)

    assert repr(r) == "<Revision: 42>"
    assert r.architectures == payload["architectures"]
    assert r.base == base
    assert r.build_url == payload["build_url"]
    assert r.confinement == payload["confinement"]
    assert r.created_at == payload["created_at"]
    assert r.revision == payload["revision"]
    assert r.sha3_384 == payload["sha3-384"]
    assert r.size == payload["size"]
    assert r.status == payload["status"]
    assert r.version == payload["version"]
    assert r.marshal() == payload


def test_releases():
    payload = {
        "revisions": [
            {
                "architectures": ["amd64"],
                "build_url": None,
                "confinement": "strict",
                "created_at": "2020-02-11T17:51:40.891996Z",
                "grade": "stable",
                "revision": 2,
                "sha3-384": "a9060ef4872ccacbfa440617a76fcd84967896b28d0d1eb7571f00a1098d766e7e93353b084ba6ad841d7b14b95ede48",
                "size": 20,
                "status": "Published",
                "version": "1.0",
            },
            {
                "architectures": ["arm64"],
                "base": "core20",
                "build_url": None,
                "confinement": "strict",
                "created_at": "2020-02-11T17:51:40.891996Z",
                "grade": "stable",
                "revision": 1,
                "sha3-384": "a9060ef4872ccacbfa440617a76fcd84967896b28d0d1eb7571f00a1098d766e7e93353b084ba6ad841d7b14b95ede48",
                "size": 20,
                "status": "Published",
                "version": "1.0",
            },
        ],
        "releases": [
            {
                "architecture": "amd64",
                "branch": None,
                "channel": "latest/stable",
                "expiration-date": None,
                "revision": 1,
                "risk": "stable",
                "track": "latest",
                "when": "2020-02-12T17:51:40.891996Z",
            },
            {
                "architecture": "amd64",
                "branch": None,
                "channel": "latest/stable",
                "expiration-date": None,
                "revision": 2,
                "risk": "stable",
                "track": "latest",
                "when": "2020-02-11T17:51:40.891996Z",
            },
            {
                "architecture": "amd64",
                "branch": None,
                "channel": "latest/edge",
                "expiration-date": None,
                "revision": 1,
                "risk": "stable",
                "track": "latest",
                "when": "2020-01-12T17:51:40.891996Z",
            },
        ],
    }

    r = releases.Releases.unmarshal(payload)

    assert repr(r) == "<Releases>"
    assert r.marshal() == payload



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

import pytest

from snapcraft.internal.meta import errors
from snapcraft.internal.meta.desktop import DesktopFile


class TestDesktopExec:
    scenarios = (
        (
            "snap name != app name",
            dict(app_name="bar", app_args="", expected_exec="foo.bar"),
        ),
        (
            "snap name != app name",
            dict(app_name="bar", app_args="--arg", expected_exec="foo.bar --arg"),
        ),
        (
            "snap name == app name",
            dict(app_name="foo", app_args="", expected_exec="foo"),
        ),
        (
            "snap name == app name",
            dict(app_name="foo", app_args="--arg", expected_exec="foo --arg"),
        ),
        (
            "snap name == app name",
            dict(app_name="foo", app_args="--arg %U", expected_exec="foo --arg %U"),
        ),
        (
            "snap name == app name",
            dict(app_name="foo", app_args="%U", expected_exec="foo %U"),
        ),
    )

    def test_generate_desktop_file(
        self, tmp_work_path, app_name, app_args, expected_exec
    ):
        snap_name = "foo"

        desktop_file_path = tmp_work_path / "app.desktop"
        with desktop_file_path.open("w") as desktop_file:
            print("[Desktop Entry]", file=desktop_file)
            print(
                "Exec={}".format(" ".join(["in-snap-exe", app_args])), file=desktop_file
            )

        d = DesktopFile(
            snap_name=snap_name,
            app_name=app_name,
            filename=desktop_file_path,
            prime_dir=tmp_work_path.as_posix(),
        )
        d.write(gui_dir=".")

        expected_desktop_file = tmp_work_path / f"{app_name}.desktop"
        assert expected_desktop_file.exists()
        with expected_desktop_file.open() as desktop_file:
            assert (
                desktop_file.read()
                == dedent(
                    """\
            [Desktop Entry]
            Exec={}

        """
                ).format(expected_exec)
            )


class TestDesktopIcon:
    scenarios = (
        (
            "icon_path preferred",
            dict(
                icon="other.png", icon_path="foo.png", expected_icon="${SNAP}/foo.png"
            ),
        ),
        (
            "icon_path with / preferred",
            dict(icon="/foo.png", icon_path="foo.png", expected_icon="${SNAP}/foo.png"),
        ),
        (
            "icon path with ${SNAP}",
            dict(
                icon="${SNAP}/foo.png", icon_path=None, expected_icon="${SNAP}/foo.png"
            ),
        ),
        ("icon name", dict(icon="foo", icon_path=None, expected_icon="foo")),
    )

    def test_generate_desktop_file(self, tmp_work_path, icon, icon_path, expected_icon):
        snap_name = app_name = "foo"

        desktop_file_path = tmp_work_path / "app.desktop"
        with desktop_file_path.open("w") as desktop_file:
            print("[Desktop Entry]", file=desktop_file)
            print("Exec=in-snap-exe", file=desktop_file)
            print("Icon={}".format(icon), file=desktop_file)

        if icon_path is not None:
            (tmp_work_path / icon_path).touch()

        d = DesktopFile(
            snap_name=snap_name,
            app_name=app_name,
            filename=desktop_file_path,
            prime_dir=tmp_work_path.as_posix(),
        )
        d.write(gui_dir=".")

        if icon_path is not None:
            d.write(icon_path=icon_path, gui_dir=".")
        else:
            d.write(gui_dir=".")

        expected_desktop_file = tmp_work_path / f"{app_name}.desktop"
        assert expected_desktop_file.exists()
        with expected_desktop_file.open() as desktop_file:
            assert (
                desktop_file.read()
                == dedent(
                    """\
            [Desktop Entry]
            Exec=foo
            Icon={}

        """
                ).format(expected_icon)
            )

    def test_generate_desktop_file_multisection(
        self, tmp_work_path, icon, icon_path, expected_icon
    ):
        snap_name = app_name = "foo"

        desktop_file_path = tmp_work_path / "app.desktop"
        with desktop_file_path.open("w") as desktop_file:
            print("[Desktop Entry]", file=desktop_file)
            print("Exec=in-snap-exe", file=desktop_file)
            print("Icon={}".format(icon), file=desktop_file)
            print("[Desktop Entry Two]", file=desktop_file)
            print("Exec=in-snap-exe2", file=desktop_file)
            print("Icon={}".format(icon), file=desktop_file)

        if icon_path is not None:
            (tmp_work_path / icon_path).touch()

        d = DesktopFile(
            snap_name=snap_name,
            app_name=app_name,
            filename=desktop_file_path,
            prime_dir=tmp_work_path.as_posix(),
        )

        if icon_path is not None:
            d.write(icon_path=icon_path, gui_dir=".")
        else:
            d.write(gui_dir=".")

        expected_desktop_file = tmp_work_path / f"{app_name}.desktop"
        assert expected_desktop_file.exists()
        with expected_desktop_file.open() as desktop_file:
            assert (
                desktop_file.read()
                == dedent(
                    """\
            [Desktop Entry]
            Exec=foo
            Icon={0}

            [Desktop Entry Two]
            Exec=foo
            Icon={0}

        """
                ).format(expected_icon)
            )


def test_not_found(tmp_path):
    with pytest.raises(errors.InvalidDesktopFileError):
        DesktopFile(
            snap_name="foo",
            app_name="foo",
            filename="desktop-file-not-found",
            prime_dir=tmp_path.as_posix(),
        )


def test_no_desktop_section(tmp_work_path):
    with open("foo.desktop", "w") as desktop_file:
        print("[Random Entry]", file=desktop_file)
        print("Exec=foo", file=desktop_file)
        print("Icon=foo", file=desktop_file)

    d = DesktopFile(
        snap_name="foo",
        app_name="foo",
        filename="foo.desktop",
        prime_dir=tmp_work_path.as_posix(),
    )

    with pytest.raises(errors.InvalidDesktopFileError):
        d.write(gui_dir=tmp_work_path.as_posix())


def test_missing_exec_entry(tmp_work_path):
    with open("foo.desktop", "w") as desktop_file:
        print("[Desktop Entry]", file=desktop_file)
        print("Icon=foo", file=desktop_file)

    d = DesktopFile(
        snap_name="foo",
        app_name="foo",
        filename="foo.desktop",
        prime_dir=tmp_work_path.as_posix(),
    )

    with pytest.raises(errors.InvalidDesktopFileError):
        d.write(gui_dir=tmp_work_path.as_posix())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018,2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import contextlib
from typing import Any, Dict, List, Optional, Set

import snapcraft.internal.project_loader._config as _config
from snapcraft.internal import errors, pluginhandler, steps

_DirtyReport = Dict[str, Dict[steps.Step, Optional[pluginhandler.DirtyReport]]]
_OutdatedReport = Dict[str, Dict[steps.Step, Optional[pluginhandler.OutdatedReport]]]


class StatusCache:
    """The StatusCache is a lazy caching interface for the status of parts."""

    def __init__(self, config: _config.Config) -> None:
        """Create a new StatusCache.

        :param _config.Config config: Project config.
        """
        self.config = config
        self._steps_run: Dict[str, Set[steps.Step]] = dict()
        self._outdated_reports: _OutdatedReport = collections.defaultdict(dict)
        self._dirty_reports: _DirtyReport = collections.defaultdict(dict)

    def should_step_run(
        self, part: pluginhandler.PluginHandler, step: steps.Step
    ) -> bool:
        """Determine if a given step of a given part should run.

        :param pluginhandler.PluginHandler part: Part in question.
        :param steps.Step step: Step in question.
        :return: Whether or not step should run.
        :rtype: bool

        A given step should run if it:
            1. Hasn't yet run
            2. Is dirty
            3. Is outdated
            4. Either (1), (2), or (3) apply to any earlier steps in the part's
               lifecycle
        """
        if (
            not self.has_step_run(part, step)
            or self.get_outdated_report(part, step) is not None
            or self.get_dirty_report(part, step) is not None
        ):
            return True

        previous_step = step.previous_step()
        if previous_step:
            return self.should_step_run(part, previous_step)

        return False

    def add_step_run(self, part: pluginhandler.PluginHandler, step: steps.Step) -> None:
        """Cache the fact that a given step has now run for the given part.

        :param pluginhandler.PluginHandler part: Part in question.
        :param steps.Step step: Step in question.
        """
        self._ensure_steps_run(part)
        self._steps_run[part.name].add(step)

    def has_step_run(self, part: pluginhandler.PluginHandler, step: steps.Step) -> bool:
        """Determine if a given step of a given part has already run.

        :param pluginhandler.PluginHandler part: Part in question.
        :param steps.Step step: Step in question.
        :return: Whether or not the step has run.
        :rtype: bool
        """
        self._ensure_steps_run(part)
        return step in self._steps_run[part.name]

    def get_outdated_report(self, part, step):
        """Obtain the outdated report for a given step of the given part.

        :param pluginhandler.PluginHandler part: Part in question.
        :param steps.Step step: Step in question.
        :return: Outdated report (could be None)
        :rtype: pluginhandler.OutdatedReport
        """
        self._ensure_outdated_report(part, step)
        return self._outdated_reports[part.name][step]

    def get_dirty_report(
        self, part: pluginhandler.PluginHandler, step: steps.Step
    ) -> Optional[pluginhandler.DirtyReport]:
        """Obtain the dirty report for a given step of the given part.

        :param pluginhandler.PluginHandler part: Part in question.
        :param steps.Step step: Step in question.
        :return: Dirty report (could be None)
        :rtype: pluginhandler.DirtyReport
        """
        self._ensure_dirty_report(part, step)
        return self._dirty_reports[part.name][step]

    def clear_step(self, part: pluginhandler.PluginHandler, step: steps.Step) -> None:
        """Clear the given step of the given part from the cache.

        :param pluginhandler.PluginHandler part: Part in question.
        :param steps.Step step: Step in question.

        This function does nothing if the step wasn't cached.
        """
        if part.name in self._steps_run:
            _remove_key(self._steps_run[part.name], step)
            if not self._steps_run[part.name]:
                _del_key(self._steps_run, part.name)
        _del_key(self._outdated_reports[part.name], step)
        if not self._outdated_reports[part.name]:
            _del_key(self._outdated_reports, part.name)
        _del_key(self._dirty_reports[part.name], step)
        if not self._dirty_reports[part.name]:
            _del_key(self._dirty_reports, part.name)

    def _ensure_steps_run(self, part: pluginhandler.PluginHandler) -> None:
        if part.name not in self._steps_run:
            self._steps_run[part.name] = _get_steps_run(part)

    def _ensure_outdated_report(
        self, part: pluginhandler.PluginHandler, step: steps.Step
    ) -> None:
        if step not in self._outdated_reports[part.name]:
            self._outdated_reports[part.name][step] = part.get_outdated_report(step)

    def _ensure_dirty_report(
        self, part: pluginhandler.PluginHandler, step: steps.Step
    ) -> None:
        # If we already have a dirty report, bail
        if step in self._dirty_reports[part.name]:
            return

        # Get the dirty report from the PluginHandler. If it's dirty, we can
        # stop here
        self._dirty_reports[part.name][step] = part.get_dirty_report(step)

        # The dirty report from the PluginHandler only takes into account
        # properties specific to that part. If it's not dirty because of those,
        # we need to expand it here to also take its dependencies (if any) into
        # account
        prerequisite_step = steps.get_dependency_prerequisite_step(step)
        dependencies = self.config.parts.get_dependencies(part.name, recursive=True)

        changed_dependencies: List[pluginhandler.Dependency] = []
        with contextlib.suppress(errors.StepHasNotRunError):
            timestamp = part.step_timestamp(step)
            for dependency in dependencies:
                # Make sure the prerequisite step of this dependency has not
                # run more recently than (or should run _before_) this step.
                try:
                    prerequisite_timestamp = dependency.step_timestamp(
                        prerequisite_step
                    )
                except errors.StepHasNotRunError:
                    dependency_changed = True
                else:
                    dependency_changed = timestamp < prerequisite_timestamp

                if dependency_changed or self.should_step_run(
                    dependency, prerequisite_step
                ):
                    changed_dependencies.append(
                        pluginhandler.Dependency(
                            part_name=dependency.name, step=prerequisite_step
                        )
                    )

            if changed_dependencies:
                self._dirty_reports[part.name][step] = pluginhandler.DirtyReport(
                    changed_dependencies=changed_dependencies
                )


def _get_steps_run(part: pluginhandler.PluginHandler) -> Set[steps.Step]:
    steps_run = set()  # type: Set[steps.Step]
    for step in steps.STEPS:
        if not part.should_step_run(step):
            steps_run.add(step)

    return steps_run


def _del_key(c: Dict[Any, Any], key: Any) -> None:
    with contextlib.suppress(KeyError):
        del c[key]


def _remove_key(c: Set[Any], key: Any) -> None:
    with contextlib.suppress(KeyError):
        c.remove(key)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
from collections import OrderedDict
from copy import deepcopy
from typing import Any, Dict, Optional, Type

from snapcraft.internal.meta.errors import PlugValidationError

logger = logging.getLogger(__name__)


class Plug:
    """Generic plug."""

    def __init__(
        self,
        *,
        plug_name: str,
        plug_dict: Optional[Dict[str, Any]] = None,
        use_string_representation: bool = False,
    ) -> None:
        self._plug_name = plug_name
        if plug_dict is None:
            self._plug_dict: Dict[str, Any] = dict()
        else:
            self._plug_dict = plug_dict

        self.use_string_representation = use_string_representation

    @property
    def plug_name(self) -> str:
        """Read-only to ensure consistency with Snap dictionary mappings."""

        return self._plug_name

    def validate(self) -> None:
        """Validate plug, raising an exception on failure."""

        # Nothing to validate (yet).
        return

    @classmethod
    def from_dict(cls, *, plug_dict: Dict[str, Any], plug_name: str) -> "Plug":
        """Create plug from dictionary."""
        interface = plug_dict.get("interface", plug_name)

        plug_class = PLUG_MAPPINGS.get(interface, None)
        if plug_class is not None:
            return plug_class.from_dict(plug_dict=plug_dict, plug_name=plug_name)

        # Handle the general case.
        plug = Plug(plug_name=plug_name)
        plug._plug_dict.update(plug_dict)
        return plug

    @classmethod
    def from_object(cls, *, plug_object: Any, plug_name: str) -> "Plug":
        if plug_object is None:
            return Plug(plug_name=plug_name)
        elif isinstance(plug_object, str):
            plug = Plug(plug_name=plug_name, use_string_representation=True)
            plug._plug_dict["interface"] = plug_object
            return plug
        elif isinstance(plug_object, dict):
            return Plug.from_dict(plug_dict=plug_object, plug_name=plug_name)

        raise RuntimeError(f"unknown syntax for plug {plug_name!r}: {plug_object!r}")

    def to_yaml_object(self) -> Any:
        # To output string short-form: "slot-name: interface-type"
        if self.use_string_representation:
            return self._plug_dict["interface"]

        # To output shortest-form: "slot-name-is-interface: <empty>"
        if not self._plug_dict:
            return None

        return OrderedDict(deepcopy(self._plug_dict))

    def __repr__(self) -> str:
        return repr(self.__dict__)

    def __str__(self) -> str:
        return str(self.__dict__)


class ContentPlug(Plug):
    """Representation of a snap content plug."""

    def __init__(
        self,
        *,
        plug_name: str,
        content: Optional[str] = None,
        default_provider: Optional[str] = None,
        target: str,
    ) -> None:
        super().__init__(plug_name=plug_name)

        self._content = content
        self._default_provider = default_provider
        self.target = target

    @property
    def interface(self) -> str:
        return "content"

    @property
    def content(self) -> str:
        if self._content:
            return self._content

        # Defaults to plug_name if unspecified.
        return self.plug_name

    @content.setter
    def content(self, content) -> None:
        self._content = content

    @property
    def provider(self) -> Optional[str]:
        if self._default_provider is None:
            return None

        if ":" in self._default_provider:
            return self._default_provider.split(":")[0]

        return self._default_provider

    def validate(self) -> None:
        if not self.target:
            raise PlugValidationError(
                plug_name=self.plug_name,
                message="`target` is required for content slot",
            )

    @classmethod
    def from_dict(cls, *, plug_dict: Dict[str, str], plug_name: str) -> "ContentPlug":
        interface = plug_dict.get("interface")
        if interface != "content":
            raise PlugValidationError(
                plug_name=plug_name,
                message="`interface={}` is invalid for content slot".format(interface),
            )

        target = plug_dict.get("target", None)
        if target is None:
            raise PlugValidationError(
                plug_name=plug_name, message="`target` is required for content slot"
            )

        return ContentPlug(
            plug_name=plug_name,
            content=plug_dict.get("content", None),
            target=target,
            default_provider=plug_dict.get("default-provider", None),
        )

    def to_yaml_object(self) -> Dict[str, str]:
        props = [("interface", self.interface)]

        # Only include content if set explicitly.
        if self._content:
            props.append(("content", self.content))

        props.append(("target", self.target))

        if self.provider:
            props.append(("default-provider", self.provider))

        return OrderedDict(props)


PLUG_MAPPINGS: Dict[str, Type[Plug]] = dict(content=ContentPlug)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016, 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import logging
import os
import subprocess
import time
from typing import BinaryIO, Tuple

from snapcraft import file_utils
from snapcraft.internal.deltas.errors import (
    DeltaFormatOptionError,
    DeltaGenerationError,
    DeltaGenerationTooBigError,
)

logger = logging.getLogger(__name__)


delta_format_options = ["xdelta3"]


class BaseDeltasGenerator:
    """Class for delta generation

    This class is responsible for the snap delta file generation
    """

    delta_size_min_pct = 90

    def __init__(
        self,
        *,
        source_path: str,
        target_path: str,
        delta_tool: str,
        delta_format: str,
        delta_file_extname: str = "delta"
    ) -> None:
        self.source_path = source_path
        self.target_path = target_path
        self.delta_format = delta_format
        self.delta_file_extname = delta_file_extname
        self.delta_tool_path = file_utils.get_snap_tool_path(delta_tool)

        # some pre-checks
        self._check_properties()
        self._check_file_existence()

    def _check_properties(self):
        if self.delta_format not in delta_format_options:
            raise DeltaFormatOptionError(
                delta_format=self.delta_format, format_options_list=delta_format_options
            )

    def _check_file_existence(self):
        if not os.path.exists(self.source_path):
            raise ValueError(
                "source file {!r} does not exist, "
                "please specify a valid source file".format(self.source_path)
            )

        if not os.path.exists(self.target_path):
            raise ValueError(
                "target file {!r} does not exist, "
                "please specify a valid target file".format(self.target_path)
            )

    def _check_delta_size_constraint(self, delta_path: str) -> None:
        """Ensure delta is sufficiently smaller than target snap.

        Although bandwidth is still saved in the case of uploading a delta
        nearly as large as the target snap, as there is a cost for delta
        reconstitution in the delta-service, pushing the full snap may be
        faster.

        Further benchmarking may help to determine the most appropriate value
        for `delta_size_min_pct`.
        """
        target_size = os.path.getsize(self.target_path)
        delta_size = os.path.getsize(delta_path)

        ratio = int((delta_size / target_size) * 100)
        if ratio >= self.delta_size_min_pct:
            raise DeltaGenerationTooBigError(
                delta_min_percentage=100 - self.delta_size_min_pct
            )

    def find_unique_file_name(self, path_hint: str) -> str:
        """Return a path on disk similar to 'path_hint' that does not exist.

        This function can be used to ensure that 'path_hint' points to a file
        on disk that does not exist. The returned filename may be a modified
        version of 'path_hint' if 'path_hint' already exists.
        """
        target = path_hint
        counter = 0
        while os.path.exists(target):
            target = "{}-{}".format(path_hint, counter)
            counter += 1
        return target

    def _setup_std_output(
        self, delta_file: str
    ) -> Tuple[str, str, BinaryIO, str, BinaryIO]:
        """Helper to setup the stdout and stderr for subprocess"""
        workdir = "/tmp/"
        _, delta_name = os.path.split(delta_file)

        stdout_path = self.find_unique_file_name(
            os.path.join(workdir, "{}.out".format(delta_name))
        )
        stdout_file = open(stdout_path, "wb")

        stderr_path = self.find_unique_file_name(
            os.path.join(workdir, "{}.err".format(delta_name))
        )
        stderr_file = open(stderr_path, "wb")

        return workdir, stdout_path, stdout_file, stderr_path, stderr_file

    def _update_progress_indicator(self, proc, progress_indicator) -> None:
        """Update the progress indicator"""
        # the caller should start the progressbar outside
        ret = None
        count = 0
        ret = proc.poll()
        while ret is None:
            if count >= progress_indicator.maxval:
                progress_indicator.start()
                count = 0
            progress_indicator.update(count)
            count += 1
            time.sleep(0.2)
            ret = proc.poll()
        print("")
        # the caller should finish the progressbar outside

    def make_delta(
        self, output_dir: str = None, progress_indicator=None, is_for_test=False
    ) -> str:
        """Call the delta generation tool to create the delta file.

        returns: generated delta file path
        """
        logger.info(
            "Generating delta for {!r}.".format(os.path.basename(self.target_path))
        )

        if output_dir is not None:
            # consider creating the delta file in the specified output_dir
            # with generated filename.
            if not os.path.exists(output_dir):
                os.makedirs(output_dir, exist_ok=True)

            _, _file_name = os.path.split(self.target_path)
            full_filename = os.path.join(output_dir, _file_name)
            delta_file = self.find_unique_file_name(
                "{}.{}".format(full_filename, self.delta_file_extname)
            )
        else:
            # create the delta file under the target_path with
            # the generated filename.
            delta_file = self.find_unique_file_name(
                "{}.{}".format(self.target_path, self.delta_file_extname)
            )

        delta_cmd = self.get_delta_cmd(self.source_path, self.target_path, delta_file)

        (
            workdir,
            stdout_path,
            stdout_file,
            stderr_path,
            stderr_file,
        ) = self._setup_std_output(delta_file)

        proc = subprocess.Popen(
            delta_cmd, stdout=stdout_file, stderr=stderr_file, cwd=workdir
        )

        if progress_indicator:
            self._update_progress_indicator(proc, progress_indicator)
        else:
            proc.wait()

        stdout_file.close()
        stderr_file.close()

        if proc.returncode != 0:
            _stdout = _stderr = ""
            with open(stdout_path) as f:
                _stdout = f.read()
            with open(stderr_path) as f:
                _stderr = f.read()

            # cleanup the testcase std logs
            if is_for_test:
                os.remove(stdout_path)
                os.remove(stderr_path)

            raise DeltaGenerationError(
                delta_format=self.delta_format,
                stdout_path=stdout_path,
                stdout=_stdout,
                stderr_path=stderr_path,
                stderr=_stderr,
                returncode=proc.returncode,
            )

        self._check_delta_size_constraint(delta_file)

        self.log_delta_file(delta_file)

        # is used for log file cleanup in unittest
        if is_for_test:
            os.remove(stdout_path)
            os.remove(stderr_path)

        return delta_file

    # ------------------------------------------------------
    # the methods need to be implemented in subclass
    # ------------------------------------------------------
    def get_delta_cmd(self, source_path, target_path, delta_file):
        """Get the delta generation command line"""
        raise NotImplementedError

    def is_returncode_unexpected(self, proc):
        """Check if the subprocess return code is expected"""
        raise NotImplementedError

    def log_delta_file(self, delta_file):
        """Log the delta generation result"""
        pass



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import stat
from textwrap import dedent

from testtools.matchers import Equals, FileContains, FileExists, Not

from snapcraft.internal.repo._base import BaseRepo, get_pkg_name_parts
from tests import unit

from . import RepoBaseTestCase


class FixXmlToolsTestCase(RepoBaseTestCase):
    def assert_fix(self, files):
        for test_file in files:
            path = test_file["path"]
            os.makedirs(os.path.dirname(path), exist_ok=True)
            with open(path, "w") as f:
                f.write(test_file["content"])

        BaseRepo.normalize("root")

        for test_file in files:
            self.assertThat(test_file["path"], FileContains(test_file["expected"]))

    def test_fix_xml2_config(self):
        self.assert_fix(
            [
                {
                    "path": os.path.join("root", "usr", "bin", "xml2-config"),
                    "content": "prefix=/usr/foo",
                    "expected": "prefix=root/usr/foo",
                }
            ]
        )

    def test_no_fix_xml2_config(self):
        self.assert_fix(
            [
                {
                    "path": os.path.join("root", "usr", "bin", "xml2-config"),
                    "content": "prefix=/foo",
                    "expected": "prefix=/foo",
                }
            ]
        )

    def test_fix_xslt_config(self):
        self.assert_fix(
            [
                {
                    "path": os.path.join("root", "usr", "bin", "xslt-config"),
                    "content": "prefix=/usr/foo",
                    "expected": "prefix=root/usr/foo",
                }
            ]
        )

    def test_no_fix_xslt_config(self):
        self.assert_fix(
            [
                {
                    "path": os.path.join("root", "usr", "bin", "xslt-config"),
                    "content": "prefix=/foo",
                    "expected": "prefix=/foo",
                }
            ]
        )

    def test_fix_xml2_xslt_config(self):
        self.assert_fix(
            [
                {
                    "path": os.path.join("root", "usr", "bin", "xml2-config"),
                    "content": "prefix=/usr/foo",
                    "expected": "prefix=root/usr/foo",
                },
                {
                    "path": os.path.join("root", "usr", "bin", "xslt-config"),
                    "content": "prefix=/usr/foo",
                    "expected": "prefix=root/usr/foo",
                },
            ]
        )

    def test_no_fix_xml2_xslt_config(self):
        self.assert_fix(
            [
                {
                    "path": os.path.join("root", "usr", "bin", "xml2-config"),
                    "content": "prefix=/foo",
                    "expected": "prefix=/foo",
                },
                {
                    "path": os.path.join("root", "usr", "bin", "xslt-config"),
                    "content": "prefix=/foo",
                    "expected": "prefix=/foo",
                },
            ]
        )


class FixShebangTestCase(RepoBaseTestCase):

    scenarios = [
        (
            "python bin dir",
            {
                "file_path": os.path.join("root", "bin", "a"),
                "content": "#!/usr/bin/python\nimport this",
                "expected": "#!/usr/bin/env python\nimport this",
            },
        ),
        (
            "python3 bin dir",
            {
                "file_path": os.path.join("root", "bin", "d"),
                "content": "#!/usr/bin/python3\nimport this",
                "expected": "#!/usr/bin/env python3\nimport this",
            },
        ),
        (
            "sbin dir",
            {
                "file_path": os.path.join("root", "sbin", "b"),
                "content": "#!/usr/bin/python\nimport this",
                "expected": "#!/usr/bin/env python\nimport this",
            },
        ),
        (
            "usr/bin dir",
            {
                "file_path": os.path.join("root", "usr", "bin", "c"),
                "content": "#!/usr/bin/python\nimport this",
                "expected": "#!/usr/bin/env python\nimport this",
            },
        ),
        (
            "usr/sbin dir",
            {
                "file_path": os.path.join("root", "usr", "sbin", "d"),
                "content": "#!/usr/bin/python\nimport this",
                "expected": "#!/usr/bin/env python\nimport this",
            },
        ),
        (
            "opt/bin dir",
            {
                "file_path": os.path.join("root", "opt", "bin", "e"),
                "content": "#!/usr/bin/python\nraise Exception()",
                "expected": "#!/usr/bin/env python\nraise Exception()",
            },
        ),
    ]

    def test_fix_shebang(self):
        for _, data in self.scenarios:
            os.makedirs(os.path.dirname(data["file_path"]), exist_ok=True)
            with open(data["file_path"], "w") as fd:
                fd.write(data["content"])

        BaseRepo.normalize("root")

        for _, data in self.scenarios:
            with open(data["file_path"], "r") as fd:
                self.expectThat(fd.read(), Equals(data["expected"]))


class RemoveUselessFilesTestCase(RepoBaseTestCase):
    def create(self, file_path):
        path = os.path.join("root", file_path)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        open(path, "w").close()

        return path

    def test_remove(self):
        paths = [
            self.create(p)
            for p in [
                os.path.join("usr", "lib", "python3.5", "sitecustomize.py"),
                os.path.join("usr", "lib", "python2.7", "sitecustomize.py"),
                os.path.join("usr", "lib", "python", "sitecustomize.py"),
            ]
        ]

        BaseRepo.normalize("root")

        for p in paths:
            self.expectThat(p, Not(FileExists()))

    def test_no_remove(self):
        path = self.create(os.path.join("opt", "python3.5", "sitecustomize.py"))

        BaseRepo.normalize("root")

        self.assertThat(path, FileExists())


class FixPkgConfigTestCase(RepoBaseTestCase):
    def test_fix_pkg_config(self):
        pc_file = os.path.join(self.tempdir, "granite.pc")

        with open(pc_file, "w") as f:
            f.write(
                dedent(
                    """\
                prefix=/usr
                exec_prefix=${prefix}
                libdir=${prefix}/lib
                includedir=${prefix}/include

                Name: granite
                Description: elementary\'s Application Framework
                Version: 0.4
                Libs: -L${libdir} -lgranite
                Cflags: -I${includedir}/granite
                Requires: cairo gee-0.8 glib-2.0 gio-unix-2.0 gobject-2.0
            """
                )
            )

        BaseRepo.normalize(self.tempdir)

        expected_pc_file_content = dedent(
            """\
            prefix={}/usr
            exec_prefix=${{prefix}}
            libdir=${{prefix}}/lib
            includedir=${{prefix}}/include

            Name: granite
            Description: elementary's Application Framework
            Version: 0.4
            Libs: -L${{libdir}} -lgranite
            Cflags: -I${{includedir}}/granite
            Requires: cairo gee-0.8 glib-2.0 gio-unix-2.0 gobject-2.0
            """.format(
                self.tempdir
            )
        )

        self.assertThat(pc_file, FileContains(expected_pc_file_content))

    def test_skip_directories_matching_pc(self):
        snap_pc_dir = os.path.join(self.tempdir, "snap.pc")

        os.mkdir(snap_pc_dir)

        # Verify the directory is not passed to fileinput, which would
        # try to file copy snap.pc to snap.pc.bak for the inplace replace.
        BaseRepo.normalize(self.tempdir)


class TestFixSymlinks(RepoBaseTestCase):
    def assert_fix(self, src, dst):
        os.makedirs("a")
        open("1", mode="w").close()

        os.symlink(src, dst)

        BaseRepo.normalize(self.tempdir)

        self.assertThat(os.readlink(dst), Equals(src))

    def test_rel_to_a(self):
        self.assert_fix("a", "rel-to-a")

    def test_abs_to_a(self):
        self.assert_fix("/a", "abs-to-a")


class FixSUIDTestCase(RepoBaseTestCase):
    def assert_mode(self, key, test_mod, expected_mod):
        file = os.path.join(self.tempdir, key)
        open(file, mode="w").close()
        os.chmod(file, test_mod)

        BaseRepo.normalize(self.tempdir)

        self.assertThat(stat.S_IMODE(os.stat(file).st_mode), Equals(expected_mod))

    def test_suid(self):
        self.assert_mode(key="suid_file", test_mod=0o4765, expected_mod=0o0765)

    def test_guid(self):
        self.assert_mode(key="guid_file", test_mod=0o2777, expected_mod=0o0777)

    def test_suid_guid(self):
        self.assert_mode(key="suid_guid_file", test_mod=0o6744, expected_mod=0o0744)

    def test_sticky_suid_guid(self):
        self.assert_mode(
            key="suid_guid_sticky_file", test_mod=0o7744, expected_mod=0o1744
        )


class TestPkgNameParts(unit.TestCase):
    def test_get_pkg_name_parts_name_only(self):
        name, version = get_pkg_name_parts("hello")
        self.assertThat(name, Equals("hello"))
        self.assertThat(version, Equals(None))

    def test_get_pkg_name_parts_all(self):
        name, version = get_pkg_name_parts("hello:i386=2.10-1")
        self.assertThat(name, Equals("hello:i386"))
        self.assertThat(version, Equals("2.10-1"))

    def test_get_pkg_name_parts_no_arch(self):
        name, version = get_pkg_name_parts("hello=2.10-1")
        self.assertThat(name, Equals("hello"))
        self.assertThat(version, Equals("2.10-1"))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017, 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Snapcraft plugins drive different build systems

Each part has a build system . Most parts are built from source using one of
a range of build systems such as CMake or Scons. Some parts are pre-built
and just copied into place, for example parts that reuse existing binary
packages.

You tell snapcraft which build system it must drive by specifying the
snapcraft plugin for that part. Every part must specify a plugin explicitly
(when you see a part that does not specify a plugin, that's because the
actual part definition is in the cloud, where the plugin is specified!)

These plugins implement a lifecycle over the following steps:

  - pull:   retrieve the source for the part from the specified location
  - build:  drive the build system determined by the choice of plugin
  - stage:  consolidate desirable files from all the parts in one tree
  - prime:  distill down to only the files which will go into the snap
  - snap:   compress the prime tree into the installable snap file

These steps correspond to snapcraft commands. So when you initiate a
'snapcraft pull' you will invoke the respective plugin for each part in
the snap, in sequence, to handle the source pull. Each part will then have a
fully populated parts/<part-name>/src/ directory. Similarly, if you then say
'snapcraft build' you will invoke the plugin responsible for each part in
turn, to build the part.

# Snapcraft Lifecycle

## Pull

In this first step, source material is retrieved from the specified
location, whether that is a URL for a tarball, a local path to a source tree
inside the snap, a revision control reference to checkout, or something
specific to the plugin such as PyPI. The plugin might also download
necessary artifacts, such as the Java SDK, which are not specific to the
particular part but which are needed by the plugin to handle its type of
build system.

All the downloaded content for each part goes into the
`parts/<part-name>/src/` directory, which acts as a cache to prevent
re-fetching content. You can clean that cache out with 'snapcraft clean'.

## Build

Snapcraft calculates an appropriate sequence to build the parts, based on
explicit 'after' references and the order of the parts in the
snapcraft.yaml. Each part is built in the `parts/<part-name>/build`
directory and installed into `parts/<part-name>/install`.

Note the install step - we might actually want to use built artifacts from
one part in the build process of another, so the `parts/<part-name>/install`
directory is useful as a 'working fresh install' of the part.

Between the plugin, the part definition YAML, and the build system of the
part, it is expected that the part can be built and installed in the right
place.

At this point you have a tree under `parts/` with a subdirectory for every
part, and underneath those, separate src, build and install trees for each
part.

## Stage

We now need to start consolidating the important pieces of each part into a
single tree. We do this twice - once in a very sweeping way that will
produce a lot of extraneous materials but is useful for debugging. This is
the 'stage' step of the lifecycle, because we move a lot of the build output
from each part into a consolidated tree under `stage/` which has the
structure of a snap but has way too much extra information.

The important thing about the staging area is that it lets you get all the
shared libraries in one place and lets you find overlapping content in the
parts. You can also try this directory as if it were a snap, and you'll have
all the debugging information in the tree, which is useful for developers.

Each part describes its own staging content - the files that should be
staged. The part will often describe "chunks" of content, called filesets,
so that they can be referred to as a useful set rather than having to call
out individual files.

## Prime

It is useful to have a directory tree which exactly mirrors the structure of
the final snap. This is the `prime/` directory, and the lifecycle includes a
'prime' step which copies only that final, required content from the
`stage/` directory into the `prime/` directory.

So the `prime/` directory contains only the content that will be put into
the final snap, unlike the staging area which may include debug and
development files not destined for your snap.

The snap metadata will also be placed in `./prime/meta` during the prime
step, so this `./prime` directory is useful for inspecting exactly what is
going into your snap or to conduct any final post-processing on snapcraft's
output.

## Snap

The final step in the snapcraft lifecycle builds a snap out of the `prime/`
directory. It will be in the top level directory, alongside snapcraft.yaml,
called <name>-<version>-<arch>.snap


# Standard part definition keywords

There are several builtin keywords which can be used in any part regardless
of the choice of plugin.

  - after: [part, part, part...]

    Snapcraft will make sure that it builds all of the listed parts before
    it tries to build this part. Essentially these listed dependencies for
    this part, useful when the part needs a library or tool built by another
    part.

    If such a dependency part is not defined in this snapcraft.yaml, it must
    be defined in the cloud parts library, and snapcraft will retrieve the
    definition of the part from the cloud. In this way, a shared library of
    parts is available to every snap author - just say 'after' and list the
    parts you want that others have already defined.

  - build-packages: [pkg, pkg, pkg...]

    A list of packages to install on the build host before building
    the part. The files from these packages typically will not go into the
    final snap unless they contain libraries that are direct dependencies of
    binaries within the snap (in which case they'll be discovered via `ldd`),
    or they are explicitly described in stage-packages.

  - stage-packages: YAML list

    A set of packages to be downloaded and unpacked to join the part
    before it's built. Note that these packages are not installed on the host.
    Like the rest of the part, all files from these packages will make it into
    the final snap unless filtered out via the `snap` keyword.

    One may simply specify packages in a flat list, in which case the packages
    will be fetched and unpacked regardless of build environment. In addition,
    a specific grammar made up of sub-lists is supported here that allows one
    to filter stage packages depending on various selectors (e.g. the target
    arch), as well as specify optional packages. The grammar is made up of two
    nestable statements: 'on' and 'try'.

    Let's discuss `on`.

        - on <selector>[,<selector>...]:
          - ...
        - else[ fail]:
          - ...

    The body of the 'on' clause is taken into account if every (AND, not OR)
    selector is true for the target build environment. Currently the only
    selectors supported are target architectures (e.g. amd64).

    If the 'on' clause doesn't match and it's immediately followed by an 'else'
    clause, the 'else' clause must be satisfied. An 'on' clause without an
    'else' clause is considered satisfied even if no selector matched. The
    'else fail' form allows erroring out if an 'on' clause was not matched.

    For example, say you only wanted to stage `foo` if building for amd64 (and
    not stage `foo` if otherwise):

        - on amd64: [foo]

    Building on that, say you wanted to stage `bar` if building on an arch
    other than amd64:

        - on amd64: [foo]
        - else: [bar]

    You can nest these for more complex behaviors:

        - on amd64: [foo]
        - else:
          - on i386: [bar]
          - on armhf: [baz]

    If your project requires a package that is only available on amd64, you can
    fail if you're not building for amd64:

        - on amd64: [foo]
        - else fail

    Now let's discuss `try`:

        - try:
          - ...
        - else:
          - ...

    The body of the 'try' clause is taken into account only when all packages
    contained within it are valid. If not, if it's immediately followed by
    'else' clauses they are tried in order, and one of them must be satisfied.
    A 'try' clause with no 'else' clause is considered satisfied even if it
    contains invalid packages.

    For example, say you wanted to stage `foo`, but it wasn't available for all
    architectures. Assuming your project builds without it, you can make it an
    optional stage package:

        - try: [foo]

    You can also add alternatives:

        - try: [foo]
        - else: [bar]

    Again, you can nest these for more complex behaviors:

        - on amd64: [foo]
        - else:
          - try: [bar]

  - organize: YAML

    Snapcraft will rename files according to this YAML sub-section. The
    content of the 'organize' section consists of old path keys, and their
    new values after the renaming.

    This can be used to avoid conflicts between parts that use the same
    name, or to map content from different parts into a common conventional
    file structure. For example:

      organize:
        usr/oldfilename: usr/newfilename
        usr/local/share/: usr/share/

    The key is the internal part filename, the value is the exposed filename
    that will be used during the staging process. You can rename whole
    subtrees of the part, or just specific files.

    Note that the path is relative (even though it is "usr/local") because
    it refers to content underneath parts/<part-name>/install which is going
    to be mapped into the stage and prime areas.

  - filesets: YAML

    When we map files into the stage and prime areas on the way to putting
    them into the snap, it is convenient to be able to refer to groups of
    files as well as individual files.  Snapcraft lets you name a fileset
    and then use it later for inclusion or exclusion of those files from the
    resulting snap.

    For example, consider man pages of header files.. You might want them
    in, or you might want to leave them out, but you definitely don't want
    to repeatedly have to list all of them either way.

    This section is thus a YAML map of fileset names (the keys) to a list of
    filenames. The list is built up by adding individual files or whole
    subdirectory paths (and all the files under that path) and wildcard
    globs, and then pruning from those paths.

    The wildcard * globs all files in that path. Exclusions are denoted by
    an initial `-`.

    For example you could add usr/local/* then remove usr/local/man/*:

      filesets:
        allbutman: [ usr/local/*, -usr/local/man/* ]
        manpages: [ usr/local/man ]

    Filenames are relative to the part install directory in
    `parts/<part-name>/install`. If you have used 'organize' to rename files
    then the filesets will be built up from the names after organization.

  - stage: YAML file and fileset list

    A list of files from a part install directory to copy into `stage/`.
    Rules applying to the list here are the same as those of filesets.
    Referencing of fileset keys is done with a $ prefixing the fileset key,
    which will expand with the value of such key.

    For example:

      stage:
        - usr/lib/*   # Everything under parts/<part-name>/install/usr/lib
        - -usr/lib/libtest.so   # Excludng libtest.so
        - $manpages             # Including the 'manpages' fileset

  - snap: YAML file and fileset list

    A list of files from a part install directory to copy into `prime/`.
    This section takes exactly the same form as the 'stage' section  but the
    files identified here will go into the ultimate snap (because the
    `prime/` directory reflects the file structure of the snap with no
    extraneous content).

  - build-attributes: [attribute1, attribute2]

    A list of special attributes that affect the build of this specific part.
    Supported attributes:

      - no-install:
        Do not run the install target provided by the plugin's build system.

        Supported by: kbuild

      - debug:
        Plugins that support the concept of build types build in Release mode
        by default. Setting the 'debug' attribute requests that they instead
        build in Debug mode.
"""

from collections import OrderedDict  # noqa

import pkg_resources  # noqa


def _get_version():
    import os as _os

    if _os.environ.get("SNAP_NAME") == "snapcraft":
        return _os.environ["SNAP_VERSION"]
    try:
        return pkg_resources.require("snapcraft")[0].version
    except pkg_resources.DistributionNotFound:
        return "devel"


# Set this early so that the circular imports aren't too painful
__version__ = _get_version()

# Workaround for potential import loops.
from snapcraft.internal import repo  # noqa isort:skip

# For backwards compatibility with external plugins.
import snapcraft._legacy_loader  # noqa: F401 isort:skip
from snapcraft.plugins.v1 import PluginV1 as BasePlugin  # noqa: F401 isort:skip
from snapcraft import common  # noqa
from snapcraft import extractors  # noqa
from snapcraft import file_utils  # noqa
from snapcraft import plugins  # noqa
from snapcraft import shell_utils  # noqa
from snapcraft import sources  # noqa

# FIXME LP: #1662658
from snapcraft._store import (  # noqa
    create_key,
    download,
    gated,
    list_keys,
    list_registered,
    login,
    register,
    register_key,
    sign_build,
    status,
    upload,
    upload_metadata,
    validate,
)

from snapcraft.project._project_options import ProjectOptions  # noqa isort:skip
